"""Tests for mutually exclusive argument groups."""

import contextlib
import dataclasses
import io
from pathlib import Path
from typing import Annotated, Literal, Optional, Tuple

import pytest
from helptext_utils import get_helptext_with_checks

import tyro
import tyro._strings


def test_required_mutex_group_basic() -> None:
    """Test basic required mutex group functionality."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True)

    def main(
        option_a: Annotated[str | None, RequiredGroup] = None,
        option_b: Annotated[int | None, RequiredGroup] = None,
    ) -> Tuple[Optional[str], Optional[int]]:
        return option_a, option_b

    # Should work with just option_a.
    assert tyro.cli(main, args=["--option-a", "hello"]) == ("hello", None)

    # Should work with just option_b.
    assert tyro.cli(main, args=["--option-b", "42"]) == (None, 42)

    # Should fail when both are provided.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--option-a", "hello", "--option-b", "42"])

    # Should fail when neither is provided.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=[])


def test_optional_mutex_group_basic() -> None:
    """Test basic optional mutex group functionality."""
    OptionalGroup = tyro.conf.create_mutex_group(required=False)

    def main(
        verbose: Annotated[bool, OptionalGroup] = False,
        quiet: Annotated[bool, OptionalGroup] = False,
    ) -> Tuple[bool, bool]:
        return verbose, quiet

    # Should work with neither option.
    assert tyro.cli(main, args=[]) == (False, False)

    # Should work with just verbose.
    assert tyro.cli(main, args=["--verbose"]) == (True, False)

    # Should work with just quiet.
    assert tyro.cli(main, args=["--quiet"]) == (False, True)

    # Should fail when both are provided.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--verbose", "--quiet"])


def test_multiple_mutex_groups() -> None:
    """Test multiple independent mutex groups in the same function."""
    GroupA = tyro.conf.create_mutex_group(required=True)
    GroupB = tyro.conf.create_mutex_group(required=False)

    def main(
        option_a1: Annotated[str | None, GroupA] = None,
        option_a2: Annotated[str | None, GroupA] = None,
        option_b1: Annotated[bool, GroupB] = False,
        option_b2: Annotated[bool, GroupB] = False,
    ) -> Tuple[Optional[str], Optional[str], bool, bool]:
        return option_a1, option_a2, option_b1, option_b2

    # Should work with one from group A and one from group B.
    assert tyro.cli(main, args=["--option-a1", "test", "--option-b1"]) == (
        "test",
        None,
        True,
        False,
    )

    # Should work with one from group A and none from group B.
    assert tyro.cli(main, args=["--option-a2", "hello"]) == (
        None,
        "hello",
        False,
        False,
    )

    # Should fail with both from group A.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--option-a1", "a", "--option-a2", "b"])

    # Should fail with both from group B.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--option-a1", "test", "--option-b1", "--option-b2"])


def test_mutex_group_with_disallow_none() -> None:
    """Test mutex groups with DisallowNone configuration."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True)

    def main(
        option_a: Annotated[str | None, RequiredGroup] = None,
        option_b: Annotated[int | None, RequiredGroup] = None,
    ) -> Tuple[Optional[str], Optional[int]]:
        return option_a, option_b

    # Regular values should work both with and without DisallowNone.
    assert tyro.cli(main, args=["--option-a", "hello"]) == ("hello", None)
    assert tyro.cli(
        main, args=["--option-a", "hello"], config=(tyro.conf.DisallowNone,)
    ) == ("hello", None)

    # DisallowNone prevents None from being passed via CLI.
    # Note: For string types, "None" is treated as the string "None", not the None value.
    # So this test validates that DisallowNone works with the mutex group feature.


def test_mutex_group_with_literal_types() -> None:
    """Test mutex groups with Literal types."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True)

    def main(
        mode: Annotated[Optional[Literal["fast", "slow"]], RequiredGroup] = None,
        threads: Annotated[int | None, RequiredGroup] = None,
    ) -> Tuple[Optional[str], Optional[int]]:
        return mode, threads

    # Should work with literal choices.
    assert tyro.cli(main, args=["--mode", "fast"]) == ("fast", None)
    assert tyro.cli(main, args=["--mode", "slow"]) == ("slow", None)

    # Should fail with invalid literal value.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--mode", "medium"])

    # Should work with the other option.
    assert tyro.cli(main, args=["--threads", "4"]) == (None, 4)


def test_mutex_group_with_path_types() -> None:
    """Test mutex groups with Path types."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True)

    def main(
        input_file: Annotated[Optional[Path], RequiredGroup] = None,
        input_dir: Annotated[Optional[Path], RequiredGroup] = None,
    ) -> Tuple[Optional[Path], Optional[Path]]:
        return input_file, input_dir

    # Should work with file path.
    result = tyro.cli(main, args=["--input-file", "/tmp/test.txt"])
    assert result == (Path("/tmp/test.txt"), None)

    # Should work with directory path.
    result = tyro.cli(main, args=["--input-dir", "/home/user"])
    assert result == (None, Path("/home/user"))


def test_mutex_group_in_dataclass() -> None:
    """Test mutex groups within dataclass fields."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True)
    OptionalGroup = tyro.conf.create_mutex_group(required=False)

    @dataclasses.dataclass
    class Config:
        # Required mutex group.
        option_a: Annotated[str | None, RequiredGroup] = None
        option_b: Annotated[int | None, RequiredGroup] = None
        # Optional mutex group.
        verbose: Annotated[bool, OptionalGroup] = False
        quiet: Annotated[bool, OptionalGroup] = False

    # Should work with valid combinations.
    config = tyro.cli(Config, args=["--option-a", "test", "--verbose"])
    assert config.option_a == "test"
    assert config.option_b is None
    assert config.verbose is True
    assert config.quiet is False

    # Should fail with invalid combinations.
    with pytest.raises(SystemExit):
        tyro.cli(Config, args=["--option-a", "a", "--option-b", "1"])

    with pytest.raises(SystemExit):
        tyro.cli(Config, args=["--option-a", "test", "--verbose", "--quiet"])


def test_mutex_group_helptext() -> None:
    """Test that mutex groups appear correctly in helptext."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True)
    OptionalGroup = tyro.conf.create_mutex_group(required=False)

    def main(
        option_a: Annotated[str | None, RequiredGroup] = None,
        option_b: Annotated[int | None, RequiredGroup] = None,
        verbose: Annotated[bool, OptionalGroup] = False,
        quiet: Annotated[bool, OptionalGroup] = False,
    ) -> None:
        """Test function with mutex groups."""
        pass

    helptext = get_helptext_with_checks(main)

    # Check for mutually exclusive section markers.
    assert "mutually exclusive" in helptext.lower()

    # Check that required group is marked as required.
    assert "required" in helptext.lower()

    # Check that options are listed.
    assert "--option-a" in helptext
    assert "--option-b" in helptext
    assert "--verbose" in helptext
    assert "--quiet" in helptext


def test_mutex_group_with_flag_create_pairs_off() -> None:
    """Test mutex groups with FlagCreatePairsOff configuration."""
    OptionalGroup = tyro.conf.create_mutex_group(required=False)

    def main(
        verbose: Annotated[bool, OptionalGroup] = False,
        quiet: Annotated[bool, OptionalGroup] = False,
    ) -> Tuple[bool, bool]:
        return verbose, quiet

    # With FlagCreatePairsOff, only positive flags should be created.
    helptext = get_helptext_with_checks(main, config=(tyro.conf.FlagCreatePairsOff,))

    # Should have positive flags.
    assert "--verbose" in helptext
    assert "--quiet" in helptext

    # Should not have negative flags.
    assert "--no-verbose" not in helptext
    assert "--no-quiet" not in helptext

    # Functionality should still work.
    assert tyro.cli(
        main, args=["--verbose"], config=(tyro.conf.FlagCreatePairsOff,)
    ) == (True, False)


def test_three_way_mutex_group() -> None:
    """Test mutex group with three options."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True)

    def main(
        option_a: Annotated[str | None, RequiredGroup] = None,
        option_b: Annotated[int | None, RequiredGroup] = None,
        option_c: Annotated[float | None, RequiredGroup] = None,
    ) -> Tuple[Optional[str], Optional[int], Optional[float]]:
        return option_a, option_b, option_c

    # Should work with any single option.
    assert tyro.cli(main, args=["--option-a", "test"]) == ("test", None, None)
    assert tyro.cli(main, args=["--option-b", "42"]) == (None, 42, None)
    assert tyro.cli(main, args=["--option-c", "3.14"]) == (None, None, 3.14)

    # Should fail with any two options.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--option-a", "a", "--option-b", "1"])

    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--option-b", "1", "--option-c", "2.0"])

    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--option-a", "a", "--option-c", "2.0"])

    # Should fail with all three options.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--option-a", "a", "--option-b", "1", "--option-c", "2.0"])


def test_mutex_group_with_defaults_not_none() -> None:
    """Test mutex groups where defaults are not None."""
    OptionalGroup = tyro.conf.create_mutex_group(required=False)

    def main(
        verbose: Annotated[bool, OptionalGroup] = False,
        verbosity_level: Annotated[int, OptionalGroup] = 0,
    ) -> Tuple[bool, int]:
        return verbose, verbosity_level

    # Should use defaults when neither is specified.
    assert tyro.cli(main, args=[]) == (False, 0)

    # Should override one default.
    assert tyro.cli(main, args=["--verbose"]) == (True, 0)
    assert tyro.cli(main, args=["--verbosity-level", "2"]) == (False, 2)

    # Should fail when both are overridden.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--verbose", "--verbosity-level", "2"])


def test_mutex_group_on_dataclass_is_ignored() -> None:
    """Test that applying a mutex group annotation to a dataclass itself is ignored.

    This documents the current limitation where mutex groups can only be applied to
    individual fields, not to composite types like dataclasses. The MutexGroup
    annotation is ignored, and the dataclass behaves normally (creating subcommands
    for Optional[Dataclass] as usual).
    """
    MutexGroup = tyro.conf.create_mutex_group(required=True)

    @dataclasses.dataclass
    class Config:
        """A configuration dataclass."""

        foo: int = 1
        bar: str = "hello"

    def main(
        config: Annotated[
            Optional[Config], MutexGroup
        ] = None,  # MutexGroup annotation is ignored.
        other: Annotated[Optional[int], MutexGroup] = None,
    ) -> Tuple[Optional[Config], Optional[int]]:
        return config, other

    # The MutexGroup annotation on the dataclass is ignored for mutual exclusion.
    # Since config is Optional[Config], it creates subcommands as normal.

    # When only --other is provided, it works.
    result = tyro.cli(main, args=["--other", "42"])
    assert result == (None, 42)

    # Both --other and config:config can be used together since the MutexGroup
    # on config is ignored (they're not actually mutually exclusive).
    # Note: order matters due to subcommand parsing.
    result = tyro.cli(main, args=["--other", "42", "config:config"])
    assert result[0] is not None
    assert result[0].foo == 1
    assert result[0].bar == "hello"
    assert result[1] == 42


def test_nested_mutex_groups() -> None:
    """Test that mutex groups work correctly across nested dataclasses."""
    SharedGroup = tyro.conf.create_mutex_group(required=False)

    @dataclasses.dataclass
    class Inner:
        """Inner config."""

        option_a: Annotated[int, SharedGroup] = 1
        option_b: Annotated[int, SharedGroup] = 2

    @dataclasses.dataclass
    class Outer:
        """Outer config."""

        inner: Inner = dataclasses.field(default_factory=Inner)
        option_c: Annotated[int, SharedGroup] = 3
        option_d: Annotated[int, SharedGroup] = 4

    # Should use defaults when nothing is specified.
    config = tyro.cli(Outer, args=[])
    assert config.inner.option_a == 1
    assert config.inner.option_b == 2
    assert config.option_c == 3
    assert config.option_d == 4

    # Should allow overriding a single option.
    config = tyro.cli(Outer, args=["--option-c", "10"])
    assert config.option_c == 10
    assert config.option_d == 4
    assert config.inner.option_a == 1
    assert config.inner.option_b == 2

    # Should allow overriding a nested option.
    config = tyro.cli(Outer, args=["--inner.option-a", "20"])
    assert config.inner.option_a == 20
    assert config.inner.option_b == 2
    assert config.option_c == 3
    assert config.option_d == 4

    # Should fail when options from the same group are overridden, even across nesting levels.
    with pytest.raises(SystemExit):
        tyro.cli(Outer, args=["--option-c", "10", "--option-d", "20"])

    with pytest.raises(SystemExit):
        tyro.cli(Outer, args=["--inner.option-a", "10", "--inner.option-b", "20"])

    # Crucially, should fail when mixing options from different nesting levels.
    with pytest.raises(SystemExit):
        tyro.cli(Outer, args=["--option-c", "10", "--inner.option-a", "20"])

    with pytest.raises(SystemExit):
        tyro.cli(Outer, args=["--option-d", "10", "--inner.option-b", "20"])


def test_mutex_group_custom_title() -> None:
    """Test that custom titles appear correctly in helptext."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True, title="output target")
    OptionalGroup = tyro.conf.create_mutex_group(
        required=False, title="verbosity level"
    )

    def main(
        option_a: Annotated[str | None, RequiredGroup] = None,
        option_b: Annotated[int | None, RequiredGroup] = None,
        verbose: Annotated[bool, OptionalGroup] = False,
        quiet: Annotated[bool, OptionalGroup] = False,
    ) -> None:
        """Test function with custom mutex group titles."""
        pass

    helptext = get_helptext_with_checks(main)

    # Check that custom titles appear in helptext.
    assert "output target" in helptext
    assert "verbosity level" in helptext

    # Check that default title does not appear.
    assert "mutually exclusive" not in helptext.lower()

    # Check that required group is still marked as required.
    assert "required" in helptext.lower()

    # Check that options are listed.
    assert "--option-a" in helptext
    assert "--option-b" in helptext
    assert "--verbose" in helptext
    assert "--quiet" in helptext

    # Verify functionality still works - mutual exclusion is enforced.
    assert tyro.cli(main, args=["--option-a", "test"]) is None
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--option-a", "test", "--option-b", "42"])
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--verbose", "--quiet"])


def test_mutex_group_custom_title_multiple_groups() -> None:
    """Test multiple independent mutex groups with different custom titles."""
    GroupA = tyro.conf.create_mutex_group(required=True, title="input source")
    GroupB = tyro.conf.create_mutex_group(required=False, title="output format")
    GroupC = tyro.conf.create_mutex_group(required=False, title="logging options")

    def main(
        input_file: Annotated[str | None, GroupA] = None,
        input_url: Annotated[str | None, GroupA] = None,
        format_json: Annotated[bool, GroupB] = False,
        format_yaml: Annotated[bool, GroupB] = False,
        log_verbose: Annotated[bool, GroupC] = False,
        log_quiet: Annotated[bool, GroupC] = False,
    ) -> None:
        """Test function with multiple custom mutex group titles."""
        pass

    helptext = get_helptext_with_checks(main)

    # Check that all custom titles appear in helptext.
    assert "input source" in helptext
    assert "output format" in helptext
    assert "logging options" in helptext

    # Check that default title does not appear.
    assert "mutually exclusive" not in helptext.lower()

    # Verify that each group works independently.
    assert tyro.cli(main, args=["--input-file", "test.txt", "--format-json"]) is None
    assert tyro.cli(main, args=["--input-url", "http://example.com"]) is None

    # Verify mutual exclusion within each group.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--input-file", "test.txt", "--input-url", "url"])
    with pytest.raises(SystemExit):
        tyro.cli(
            main, args=["--input-file", "test.txt", "--format-json", "--format-yaml"]
        )
    with pytest.raises(SystemExit):
        tyro.cli(
            main, args=["--input-file", "test.txt", "--log-verbose", "--log-quiet"]
        )


def test_mutex_error_message_format() -> None:
    """Test that mutex error messages show clean argument names, not internal representations."""
    RequiredGroup = tyro.conf.create_mutex_group(required=True, title="output target")
    OptionalGroup = tyro.conf.create_mutex_group(
        required=False, title="verbosity level"
    )

    def main(
        target_stream: Annotated[
            Optional[Literal["stdout", "stderr"]], RequiredGroup
        ] = None,
        target_file: Annotated[Optional[Path], RequiredGroup] = None,
        verbose: Annotated[bool, OptionalGroup] = False,
        very_verbose: Annotated[bool, OptionalGroup] = False,
    ) -> None:
        """Demonstrate mutually exclusive argument groups."""
        pass

    # Test required mutex group error message.
    stderr = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stderr(stderr):
            tyro.cli(
                main,
                args=["--target-stream", "stdout", "--target-file", "/tmp/output.txt"],
                config=(tyro.conf.DisallowNone, tyro.conf.FlagCreatePairsOff),
            )

    error_message = tyro._strings.strip_ansi_sequences(stderr.getvalue())
    # Should show clean argument names.
    assert "--target-stream" in error_message
    assert "--target-file" in error_message
    # Both backends have different error message wording but both should indicate mutual exclusion.
    # argparse: "not allowed with argument"
    # tyro: "not allowed together"
    assert "not allowed" in error_message
    # Should NOT show internal ArgumentDefinition representation.
    assert "ArgumentDefinition" not in error_message
    assert "intern_prefix" not in error_message
    assert "extern_prefix" not in error_message

    # Test optional mutex group error message.
    stderr = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stderr(stderr):
            tyro.cli(
                main,
                args=["--target-stream", "stdout", "--verbose", "--very-verbose"],
                config=(tyro.conf.DisallowNone, tyro.conf.FlagCreatePairsOff),
            )

    error_message = tyro._strings.strip_ansi_sequences(stderr.getvalue())
    # Should show clean argument names.
    assert "--verbose" in error_message
    assert "--very-verbose" in error_message
    # Both backends should indicate mutual exclusion.
    assert "not allowed" in error_message
    # Should NOT show internal ArgumentDefinition representation.
    assert "ArgumentDefinition" not in error_message


def test_mutex_group_with_short_alias() -> None:
    """Test mutex groups with one-letter aliases.

    Adapted from Aleksander Krastev: https://github.com/brentyi/tyro/issues/419
    Short flags (like -f) should not be inverted to --no-f.
    """
    MutexGroup = tyro.conf.create_mutex_group(required=False)

    @dataclasses.dataclass
    class Config:
        foo: Annotated[
            bool,
            MutexGroup,
            tyro.conf.FlagCreatePairsOff,
            tyro.conf.arg(aliases=["-f"]),
        ] = False

    # Should work with short alias.
    config = tyro.cli(Config, args=["-f"])
    assert config.foo is True

    # Should work with long flag.
    config = tyro.cli(Config, args=["--foo"])
    assert config.foo is True

    # Should work with no flag.
    config = tyro.cli(Config, args=[])
    assert config.foo is False


def test_boolean_flag_with_short_alias_helptext() -> None:
    """Test that boolean flags with short aliases show correct help text.

    Adapted from Aleksander Krastev: https://github.com/brentyi/tyro/issues/419
    Help text should show -f, --foo, --no-foo, not -f, --no-, --foo, --no-foo.
    """

    @dataclasses.dataclass
    class Config:
        foo: Annotated[bool, tyro.conf.arg(aliases=["-f"])] = False

    helptext = get_helptext_with_checks(Config)

    # Should have all valid flags.
    assert "-f" in helptext
    assert "--foo" in helptext
    assert "--no-foo" in helptext

    # Should NOT have malformed --no- flag.
    # The malformed flag would show up as "--no-," (with comma) or "--no- " (with space).
    assert "--no-," not in helptext
    assert "--no- " not in helptext


def test_boolean_flag_with_short_alias_functionality() -> None:
    """Test that boolean flags with short aliases work correctly.

    Adapted from Aleksander Krastev: https://github.com/brentyi/tyro/issues/419
    """

    @dataclasses.dataclass
    class Config:
        foo: Annotated[bool, tyro.conf.arg(aliases=["-f"])] = False

    # Short alias should set to True.
    config = tyro.cli(Config, args=["-f"])
    assert config.foo is True

    # Long flag should set to True.
    config = tyro.cli(Config, args=["--foo"])
    assert config.foo is True

    # Negated long flag should set to False.
    config = tyro.cli(Config, args=["--no-foo"])
    assert config.foo is False

    # No flag should use default.
    config = tyro.cli(Config, args=[])
    assert config.foo is False


def test_short_multi_char_alias() -> None:
    """Test that -foo (short flag with multiple chars) is not inverted.

    Short flags (starting with single -) should never be inverted, regardless
    of the number of characters after the dash.
    """

    @dataclasses.dataclass
    class Config:
        bar: Annotated[bool, tyro.conf.arg(aliases=["-foo"])] = False

    # Short multi-char alias should work.
    config = tyro.cli(Config, args=["-foo"])
    assert config.bar is True

    # Long flag should work.
    config = tyro.cli(Config, args=["--bar"])
    assert config.bar is True

    # Negated long flag should work.
    config = tyro.cli(Config, args=["--no-bar"])
    assert config.bar is False

    # Default should work.
    config = tyro.cli(Config, args=[])
    assert config.bar is False


def test_short_multi_char_alias_helptext() -> None:
    """Test that -foo alias shows correct help text without --no-foo."""

    @dataclasses.dataclass
    class Config:
        bar: Annotated[bool, tyro.conf.arg(aliases=["-foo"])] = False

    helptext = get_helptext_with_checks(Config)

    # Should have valid flags.
    assert "-foo" in helptext
    assert "--bar" in helptext
    assert "--no-bar" in helptext

    # Should NOT have --no-foo (short flags are not inverted).
    assert "--no-foo" not in helptext


def test_short_multi_char_alias_in_mutex_group() -> None:
    """Test -foo alias in mutex group works correctly."""
    MutexGroup = tyro.conf.create_mutex_group(required=False)

    @dataclasses.dataclass
    class Config:
        bar: Annotated[
            bool,
            MutexGroup,
            tyro.conf.FlagCreatePairsOff,
            tyro.conf.arg(aliases=["-foo"]),
        ] = False

    # Should work with short multi-char alias.
    config = tyro.cli(Config, args=["-foo"])
    assert config.bar is True

    # Should work with long flag.
    config = tyro.cli(Config, args=["--bar"])
    assert config.bar is True

    # Default should work.
    config = tyro.cli(Config, args=[])
    assert config.bar is False


def test_long_single_char_alias() -> None:
    """Test that --f (long flag with single char) IS inverted to --no-f.

    Long flags (starting with --) should always be inverted, regardless
    of the number of characters after the dashes.
    """

    @dataclasses.dataclass
    class Config:
        bar: Annotated[bool, tyro.conf.arg(aliases=["--f"])] = False

    # Long single-char alias should work.
    config = tyro.cli(Config, args=["--f"])
    assert config.bar is True

    # Long flag should work.
    config = tyro.cli(Config, args=["--bar"])
    assert config.bar is True

    # Negated single-char alias should work.
    config = tyro.cli(Config, args=["--no-f"])
    assert config.bar is False

    # Negated long flag should work.
    config = tyro.cli(Config, args=["--no-bar"])
    assert config.bar is False

    # Default should work.
    config = tyro.cli(Config, args=[])
    assert config.bar is False


def test_long_single_char_alias_helptext() -> None:
    """Test that --f alias shows --no-f in help text."""

    @dataclasses.dataclass
    class Config:
        bar: Annotated[bool, tyro.conf.arg(aliases=["--f"])] = False

    helptext = get_helptext_with_checks(Config)

    # Should have all valid flags including --no-f.
    assert "--f" in helptext
    assert "--no-f" in helptext
    assert "--bar" in helptext
    assert "--no-bar" in helptext


def test_long_single_char_alias_in_mutex_group() -> None:
    """Test --f alias in mutex group works correctly."""
    MutexGroup = tyro.conf.create_mutex_group(required=False)

    @dataclasses.dataclass
    class Config:
        bar: Annotated[
            bool,
            MutexGroup,
            tyro.conf.FlagCreatePairsOff,
            tyro.conf.arg(aliases=["--f"]),
        ] = False

    # Should work with long single-char alias.
    config = tyro.cli(Config, args=["--f"])
    assert config.bar is True

    # Should work with long flag.
    config = tyro.cli(Config, args=["--bar"])
    assert config.bar is True

    # Default should work.
    config = tyro.cli(Config, args=[])
    assert config.bar is False



"""Overriding YAML configs

:mod:`tyro` understands a wide range of data structures, including standard
dictionaries and lists.

If you have a library of existing YAML files that you want to use,
:func:`tyro.cli` can help override values within them.

.. note::

    We recommend dataclass configs for new projects.

Usage:

    python ./02_overriding_yaml.py --help
    python ./02_overriding_yaml.py --training.checkpoint-steps 300 1000 9000
"""

import yaml

import tyro

# YAML configuration. This could also be loaded from a file! Environment
# variables are an easy way to select between different YAML files.
default_yaml = r"""
exp_name: test
optimizer:
    learning_rate: 0.0001
    type: adam
training:
    batch_size: 32
    num_steps: 10000
    checkpoint_steps:
    - 500
    - 1000
    - 1500
""".strip()

if __name__ == "__main__":
    # Convert our YAML config into a nested dictionary.
    default_config = yaml.safe_load(default_yaml)

    # Override fields in the dictionary.
    overridden_config = tyro.cli(dict, default=default_config)

    # Print the overridden config.
    overridden_yaml = yaml.safe_dump(overridden_config)
    print(overridden_yaml)



import contextlib
import dataclasses
import enum
import io
import json
import os
import pathlib
from typing import (
    Annotated,
    Any,
    Dict,
    Generic,
    List,
    Literal,
    NotRequired,
    Optional,
    Tuple,
    TypedDict,
    TypeVar,
    cast,
)

import pytest
from helptext_utils import get_helptext_with_checks

import tyro
import tyro._strings


def test_helptext() -> None:
    @dataclasses.dataclass
    class Helptext:
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: Annotated[int, "ignored"]

        z: int = 3
        """Documentation 3"""

    helptext = get_helptext_with_checks(Helptext)
    assert cast(str, helptext) in helptext
    assert "x INT" in helptext
    assert "y INT" in helptext
    assert "z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_helptext_sphinx_autodoc_style() -> None:
    @dataclasses.dataclass
    class Helptext:
        """This docstring should be printed as a description."""

        x: int  #: Documentation 1

        #:Documentation 2
        y: Annotated[int, "ignored"]
        z: int = 3

    helptext = get_helptext_with_checks(Helptext)
    assert cast(str, helptext) in helptext
    assert "x INT" in helptext
    assert "y INT" in helptext
    assert "z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert ": Documentation 1" not in helptext
    assert "Documentation 2 (required)" in helptext
    assert ":Documentation 2" not in helptext

    # :Documentation 2 should not be applied to `z`.
    assert helptext.count("Documentation 2") == 1


def test_helptext_from_class_docstring() -> None:
    @dataclasses.dataclass
    class Helptext2:
        """This docstring should be printed as a description.

        Attributes:
            x: Documentation 1
            y: Documentation 2
            z: Documentation 3
        """

        x: int
        y: Annotated[int, "ignored"]
        z: int = 3

    helptext = get_helptext_with_checks(Helptext2)
    assert "This docstring should be printed as a description" in helptext
    assert "Attributes" not in helptext
    assert "x INT" in helptext
    assert "y INT" in helptext
    assert "z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_helptext_from_class_docstring_args() -> None:
    @dataclasses.dataclass
    class Helptext3:
        """This docstring should be printed as a description.

        Args:
            x: Documentation 1
            y: Documentation 2
            z: Documentation 3
        """

        x: int
        y: Annotated[int, "ignored"]
        z: int = 3

    helptext = get_helptext_with_checks(Helptext3)
    assert "This docstring should be printed as a description" in helptext
    assert "Args" not in helptext
    assert "x INT" in helptext
    assert "y INT" in helptext
    assert "z INT" in helptext
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2 (required)" in helptext
    assert "Documentation 3 (default: 3)" in helptext


def test_helptext_inherited() -> None:
    class UnrelatedParentClass:
        pass

    @dataclasses.dataclass
    class ActualParentClass:
        """This docstring should be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        # fmt: off

        z: int = 3
        def some_method(self) -> None:  # noqa
            """Coverage stress test."""
        # fmt: on

    @dataclasses.dataclass
    class ChildClass(UnrelatedParentClass, ActualParentClass):
        pass

    helptext = get_helptext_with_checks(ChildClass)
    assert "Documentation 1" in helptext
    assert "Documentation 2" in helptext


def test_helptext_inherited_default_override() -> None:
    @dataclasses.dataclass
    class ParentClass:
        """This docstring should __not__ be printed as a description."""

        x: int  # Documentation 1

        # Documentation 2
        y: int

        # fmt: off

        z: int = 3
        def some_method(self) -> None:  # noqa
            """Coverage stress test."""
        # fmt: on

    @dataclasses.dataclass
    class ChildClass(ParentClass):
        """This docstring should be printed as a description."""

    def main(x: ParentClass = ChildClass(x=5, y=5)) -> Any:
        """Main function."""
        return x

    helptext = get_helptext_with_checks(main)
    assert "Main function." in helptext
    assert "Documentation 1" in helptext
    assert "Documentation 2" in helptext
    assert "__not__" not in helptext
    assert helptext.count("should be printed") == 1


def test_helptext_nested() -> None:
    """For nested classes, we should pull helptext from the outermost docstring if
    possible. The class docstring can be used as a fallback."""

    class Inner:
        def __init__(self, a: int):
            """Something

            Args:
                a (int): Hello world!
            """

    def main_with_docstring(a: Inner) -> None:
        """main_with_docstring.

        Args:
            a: Documented in function.
        """

    def main_no_docstring(a: Inner) -> None:
        """main_no_docstring."""

    helptext = get_helptext_with_checks(main_with_docstring)
    assert "Documented in function" in helptext and str(Inner.__doc__) not in helptext
    assert "main_with_docstring." in helptext
    assert "Args:" not in helptext
    assert "Args:" not in helptext
    assert "Hello world!" in helptext

    helptext = get_helptext_with_checks(main_no_docstring)
    assert "Something" in helptext
    assert "main_no_docstring." in helptext
    assert "Args:" not in helptext
    assert "Hello world!" in helptext


def test_helptext_defaults() -> None:
    class Color(enum.Enum):
        RED = enum.auto()
        GREEN = enum.auto()
        BLUE = enum.auto()

    @dataclasses.dataclass
    class HelptextWithVariousDefaults:
        x: pathlib.Path = pathlib.Path("/some/path/to/a/file")
        y: Color = Color.RED
        z: str = "%"

    helptext = get_helptext_with_checks(HelptextWithVariousDefaults)
    assert "show this help message and exit" in helptext
    assert "--x PATH" in helptext
    assert "(default: /some/path/to/a/file)" in helptext
    assert "--y {RED,GREEN,BLUE}" in helptext
    assert "(default: RED)" in helptext
    assert "--z STR" in helptext
    assert "(default: %)" in helptext


def test_helptext_upath_defaults() -> None:
    """Test UPath helptext with metavar and default rendering."""
    from upath import UPath

    @dataclasses.dataclass
    class HelptextWithUPath:
        x: UPath = UPath("s3://bucket/path")

    helptext = get_helptext_with_checks(HelptextWithUPath)
    assert "--x UPATH" in helptext
    assert "(default: s3://bucket/path)" in helptext


def test_multiline_helptext() -> None:
    @dataclasses.dataclass
    class HelptextMultiline:
        x: int  # Documentation 1

        # This comment should be ignored!

        # Documentation 2
        # Next line of documentation 2
        y: int

        z: int = 3
        """Documentation 3
        Next line of documentation 3"""

    helptext = get_helptext_with_checks(HelptextMultiline)
    assert "Documentation 1 (required)" in helptext
    assert "Documentation 2" in helptext
    assert "documentation 2" in helptext
    assert "Documentation 3" in helptext
    assert "documentation 3" in helptext


def test_grouped_helptext() -> None:
    @dataclasses.dataclass
    class HelptextGrouped:
        x: int  # Documentation 1
        # Description of both y and z.
        y: int
        z: int = 3

    helptext = get_helptext_with_checks(HelptextGrouped)
    assert "Documentation 1 (required)" in helptext
    assert "Description of both y and z. (required)" in helptext
    assert "Description of both y and z. (default: 3)" in helptext


def test_none_default_value_helptext() -> None:
    @dataclasses.dataclass
    class Config:
        x: Optional[int] = None
        """An optional variable."""

    helptext = get_helptext_with_checks(Config)
    assert "--x {None}|INT" in helptext
    assert "An optional variable. (default: None)" in helptext


def test_helptext_hard_bool() -> None:
    @dataclasses.dataclass
    class HelptextHardString:
        # fmt: off
        x: bool = (
            False
        )
        """Helptext. 2% milk."""
        # fmt: on

    # The percent symbol needs some extra handling in argparse.
    # https://stackoverflow.com/questions/21168120/python-argparse-errors-with-in-help-string

    helptext = get_helptext_with_checks(HelptextHardString)
    assert "--x" in helptext
    assert "2% milk." in helptext


def test_helptext_with_inheritance() -> None:
    @dataclasses.dataclass
    class Parent:
        # fmt: off
        x: str = (
            "This docstring may be tougher to parse!"
        )
        """Helptext."""
        # fmt: on

    @dataclasses.dataclass
    class Child(Parent):
        pass

    helptext = get_helptext_with_checks(Child)
    assert "--x STR" in helptext
    assert "Helptext." in helptext
    assert "(default: 'This docstring" in helptext


def test_helptext_with_inheritance_overriden() -> None:
    @dataclasses.dataclass
    class Parent2:
        # fmt: off
        x: str = (
            "This docstring may be tougher to parse!"
        )
        """Helptext."""
        # fmt: on

    @dataclasses.dataclass
    class Child2(Parent2):
        # fmt: off
        x: str = (
            "This docstring may be tougher to parse?"
        )
        """Helptext!"""
        # fmt: on

    helptext = get_helptext_with_checks(Child2)
    assert "--x STR" in helptext
    assert "Helptext! (default: 'This" in helptext


def test_tuple_helptext() -> None:
    @dataclasses.dataclass
    class TupleHelptext:
        x: Tuple[int, str, float]

    helptext = get_helptext_with_checks(TupleHelptext)
    assert "--x INT STR FLOAT" in helptext


def test_tuple_helptext_defaults() -> None:
    @dataclasses.dataclass
    class TupleHelptextDefaults:
        x: Tuple[int, str, str] = (5, "hello world", "hello")

    helptext = get_helptext_with_checks(TupleHelptextDefaults)
    assert "--x INT STR STR" in helptext
    assert "(default: 5 'hello world' hello)" in helptext


def test_generic_helptext() -> None:
    T = TypeVar("T")

    @dataclasses.dataclass
    class GenericTupleHelptext(Generic[T]):
        x: T

    helptext = get_helptext_with_checks(GenericTupleHelptext[int])
    assert "--x INT" in helptext


def test_generic_tuple_helptext() -> None:
    T = TypeVar("T")

    @dataclasses.dataclass
    class GenericTupleHelptext(Generic[T]):
        x: Tuple[T, T, T]

    helptext = get_helptext_with_checks(GenericTupleHelptext[int])
    assert "--x INT INT INT" in helptext


def test_generic_list_helptext() -> None:
    T = TypeVar("T")

    @dataclasses.dataclass
    class GenericTupleHelptext(Generic[T]):
        x: List[T]

    helptext = get_helptext_with_checks(GenericTupleHelptext[int])
    assert "--x [INT [INT ...]]" in helptext


def test_literal_helptext() -> None:
    @dataclasses.dataclass
    class LiteralHelptext:
        x: Literal[1, 2, 3]
        """A number."""

    helptext = get_helptext_with_checks(LiteralHelptext)
    assert "--x {1,2,3}" in helptext
    assert "A number. (required)" in helptext


def test_optional_literal_helptext() -> None:
    @dataclasses.dataclass
    class OptionalLiteralHelptext:
        x: Optional[Literal[1, 2, 3]] = None
        """A number."""

    helptext = get_helptext_with_checks(OptionalLiteralHelptext)
    assert "--x {None,1,2,3}" in helptext
    assert "A number. (default: None)" in helptext


def test_multiple_subparsers_helptext() -> None:
    @dataclasses.dataclass
    class Subcommand1:
        """2% milk."""  # % symbol is prone to bugs in argparse.

        x: int = 0

    @dataclasses.dataclass
    class Subcommand2:
        y: int = 1

    @dataclasses.dataclass
    class Subcommand3:
        z: int = 2

    @dataclasses.dataclass
    class MultipleSubparsers:
        # Field a description.
        a: Subcommand1 | Subcommand2 | Subcommand3
        # Field b description.
        b: Subcommand1 | Subcommand2 | Subcommand3
        # Field c description.
        c: Subcommand1 | Subcommand2 | Subcommand3 = dataclasses.field(
            default_factory=Subcommand3
        )

        d: bool = False

    helptext = get_helptext_with_checks(MultipleSubparsers)

    assert "2% milk." in helptext
    assert "Field a description." in helptext
    # With tyro backend, all subparser groups in the frontier are shown.
    # With argparse backend, only the first subparser group is shown.

    if tyro._experimental_options["backend"] == "tyro":
        assert "Field b description." in helptext
        assert "Field c description." in helptext
    else:
        assert "Field b description." not in helptext
        assert "Field c description." not in helptext

    # Not enough args for usage shortening to kick in.
    assert "[OPTIONS]" not in helptext
    assert "[B:SUBCOMMAND2 OPTIONS]" not in helptext

    helptext = get_helptext_with_checks(
        MultipleSubparsers,
        args=["a:subcommand1", "b:subcommand1", "--help"],
    )

    assert "2% milk." in helptext
    assert "Field a description." not in helptext
    assert "Field b description." not in helptext
    assert "Field c description." in helptext
    assert "--no-d" not in helptext
    assert "(default: c:subcommand3)" in helptext

    # Not enough args for usage shortening to kick in.
    assert "[OPTIONS]" not in helptext
    assert "[B:SUBCOMMAND1 OPTIONS]" not in helptext
    assert "[B:SUBCOMMAND2 OPTIONS]" not in helptext

    # Argument should be pushed to the leaf.
    assert "--d, --no-d" in get_helptext_with_checks(
        MultipleSubparsers,
        args=["a:subcommand1", "b:subcommand1", "c:subcommand2", "--help"],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )


def test_multiple_subparsers_helptext_shortened_usage() -> None:
    @dataclasses.dataclass
    class Subcommand1:
        """2% milk."""  # % symbol is prone to bugs in argparse.

        aaaaaaa: int = 0
        bbbbbbb: int = 0
        ccccccc: int = 0
        ddddddd: int = 0
        eeeeeee: int = 0

    @dataclasses.dataclass
    class Subcommand2:
        aaaaaaa: int = 0
        bbbbbbb: int = 0
        ccccccc: int = 0
        ddddddd: int = 0
        eeeeeee: int = 0

    @dataclasses.dataclass
    class Subcommand3:
        a: int = 0
        b: int = 0
        c: int = 0
        d: int = 0
        e: int = 0

    @dataclasses.dataclass
    class MultipleSubparsers:
        # Field a description.
        a: Subcommand1 | Subcommand2 | Subcommand3
        # Field b description.
        b: Subcommand1 | Subcommand2 | Subcommand3
        # Field c description.
        c: Subcommand1 | Subcommand2 | Subcommand3 = dataclasses.field(
            default_factory=Subcommand3
        )

        ddddddd: bool = False
        fffffff: bool = False
        ggggggg: bool = False
        hhhhhhh: bool = False
        iiiiiii: bool = False

    helptext = get_helptext_with_checks(MultipleSubparsers)

    assert "2% milk." in helptext
    assert "Field a description." in helptext
    # With tyro backend, all subparser groups in the frontier are shown.
    # With argparse backend, only the first subparser group is shown.

    if tyro._experimental_options["backend"] == "tyro":
        assert "Field b description." in helptext
        assert "Field c description." in helptext
    else:
        assert "Field b description." not in helptext
        assert "Field c description." not in helptext

    assert "[OPTIONS]" in helptext
    assert "[B:SUBCOMMAND2 OPTIONS]" not in helptext

    helptext = get_helptext_with_checks(
        MultipleSubparsers, args=["a:subcommand1", "b:subcommand1", "--help"]
    )

    assert "2% milk." in helptext
    assert "Field a description." not in helptext
    assert "Field b description." not in helptext
    assert "Field c description." in helptext
    assert "(default: c:subcommand3)" in helptext

    assert "[OPTIONS]" not in helptext
    assert "[B:SUBCOMMAND1 OPTIONS]" in helptext
    assert "[B:SUBCOMMAND2 OPTIONS]" not in helptext


def test_optional_helptext() -> None:
    @dataclasses.dataclass
    class OptionalHelptext:
        """This docstring should be printed as a description. 2% milk."""

        x: Optional[int]  # Documentation 1

        # Documentation 2
        y: List[Optional[int]]

        z: Optional[int] = 3
        """Documentation 3"""

    helptext = get_helptext_with_checks(OptionalHelptext)
    assert cast(str, cast(str, OptionalHelptext.__doc__)[:20]) in helptext
    assert "2% milk" in helptext
    assert "--x {None}|INT" in helptext
    assert "--y [{None}|INT [{None}|INT ...]]" in helptext
    assert "[--z {None}|INT]" in helptext


def test_metavar_0() -> None:
    def main(x: Literal[0, 1, 2, 3] | Tuple[int, int]) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3}|{INT INT}" in helptext


def test_metavar_1() -> None:
    def main(
        x: Literal[0, 1, 2, 3] | Literal["hey,there", "hello"] | List[int],
    ) -> None:
        pass

    # The comma formatting is unfortunate, but matches argparse's default behavior.
    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3,hey,there,hello}|{[INT [INT ...]]}" in helptext


def test_metavar_2() -> None:
    def main(
        x: Tuple[
            Literal[0, 1, 2, 3],
            int | str,
        ],
    ) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3} INT|STR" in helptext


def test_metavar_3() -> None:
    def main(
        x: Literal[0, 1, 2, 3] | Tuple[int, int] | Tuple[str],
    ) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3}|{INT INT}|STR" in helptext


def test_metavar_4() -> None:
    def main(
        x: Literal[0, 1, 2, 3] | Tuple[int, int] | Tuple[str, str, str] | Literal[True],
    ) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x {0,1,2,3}|{INT INT}|{STR STR STR}|{True}" in helptext


def test_metavar_5() -> None:
    def main(
        x: List[Tuple[int, int] | Tuple[str, str]] = [(1, 1), (2, 2)],
    ) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "[--x [{INT INT}|{STR STR} [{INT INT}|{STR STR} ...]]]" in helptext


def test_metavar_6() -> None:
    def main(x: Dict[Tuple[int, int] | Tuple[str, str], Tuple[int, int]]) -> dict:
        return x

    helptext = get_helptext_with_checks(main)
    assert (
        "--x [{INT INT}|{STR STR} INT INT [{INT INT}|{STR STR} INT INT ...]]"
        in helptext
    )


def test_empty_metavar() -> None:
    """Test that empty metavar doesn't add trailing space."""

    @dataclasses.dataclass
    class Args:
        verbose: Annotated[int, tyro.conf.arg(metavar="", aliases=("-v",))] = 0

    helptext = get_helptext_with_checks(Args)
    # Should have no space after comma: "-v, --verbose".
    # Not: "-v , --verbose".
    assert "-v, --verbose" in helptext
    assert "-v , --verbose" not in helptext


def test_comment_in_subclass_list() -> None:
    @dataclasses.dataclass
    class Something(
        # This text should not show up in the helptext anywhere.
        object,
    ):
        a: int

        # But this text should!
        b: int

    helptext = get_helptext_with_checks(Something)
    assert "This text should not" not in helptext
    assert "But this text should!" in helptext


def test_pathlike() -> None:
    def main(x: os.PathLike) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--x PATHLIKE " in helptext


def test_nested_bool() -> None:
    @dataclasses.dataclass
    class Child:
        x: bool = False

    def main(child: Child) -> None:
        pass

    helptext = get_helptext_with_checks(main)
    assert "--child.x | --child.no-x" in helptext


def test_multiple_subparsers_helptext_hyphens() -> None:
    @dataclasses.dataclass
    class SubcommandOne:
        """2% milk."""  # % symbol is prone to bugs in argparse.

        arg_x: int = 0
        arg_flag: bool = False

    @dataclasses.dataclass
    class SubcommandTwo:
        arg_y: int = 1

    @dataclasses.dataclass
    class SubcommandThree:
        arg_z: int = 2

    @dataclasses.dataclass
    class MultipleSubparsers:
        # Field a description.
        a: SubcommandOne | SubcommandTwo | SubcommandThree
        # Field b description.
        b: SubcommandOne | SubcommandTwo | SubcommandThree
        # Field c description.
        c: SubcommandOne | SubcommandTwo | SubcommandThree = dataclasses.field(
            default_factory=SubcommandThree
        )

    helptext = get_helptext_with_checks(MultipleSubparsers)

    assert "2% milk." in helptext
    assert "Field a description." in helptext
    # With tyro backend, all subparser groups in the frontier are shown.
    # With argparse backend, only the first subparser group is shown.

    if tyro._experimental_options["backend"] == "tyro":
        assert "Field b description." in helptext
        assert "Field c description." in helptext
    else:
        assert "Field b description." not in helptext
        assert "Field c description." not in helptext

    helptext = get_helptext_with_checks(
        MultipleSubparsers, args=["a:subcommand-one", "b:subcommand-one", "--help"]
    )

    assert "2% milk." in helptext
    assert "Field a description." not in helptext
    assert "Field b description." not in helptext
    assert "Field c description." in helptext
    assert "(default: c:subcommand-three)" in helptext
    assert "--b.arg-x" in helptext
    assert "--b.no-arg-flag" in helptext
    assert "--b.arg-flag" in helptext


def test_multiple_subparsers_helptext_underscores() -> None:
    @dataclasses.dataclass
    class SubcommandOne:
        """2% milk."""  # % symbol is prone to bugs in argparse.

        arg_x: int = 0
        arg_flag: bool = False

    @dataclasses.dataclass
    class SubcommandTwo:
        arg_y: int = 1

    @dataclasses.dataclass
    class SubcommandThree:
        arg_z: int = 2

    @dataclasses.dataclass
    class MultipleSubparsers:
        # Field a description.
        a: SubcommandOne | SubcommandTwo | SubcommandThree
        # Field b description.
        b: SubcommandOne | SubcommandTwo | SubcommandThree
        # Field c description.
        c: SubcommandOne | SubcommandTwo | SubcommandThree = dataclasses.field(
            default_factory=SubcommandThree
        )

    helptext = get_helptext_with_checks(MultipleSubparsers, use_underscores=True)

    assert "2% milk." in helptext
    assert "Field a description." in helptext
    # With tyro backend, all subparser groups in the frontier are shown.
    # With argparse backend, only the first subparser group is shown.

    if tyro._experimental_options["backend"] == "tyro":
        assert "Field b description." in helptext
        assert "Field c description." in helptext
    else:
        assert "Field b description." not in helptext
        assert "Field c description." not in helptext

    helptext = get_helptext_with_checks(
        MultipleSubparsers,
        args=["a:subcommand_one", "b:subcommand_one", "--help"],
        use_underscores=True,
    )

    assert "2% milk." in helptext
    assert "Field a description." not in helptext
    assert "Field b description." not in helptext
    assert "Field c description." in helptext
    assert "(default: c:subcommand_three)" in helptext
    assert "--b.arg_x" in helptext
    assert "--b.no_arg_flag" in helptext
    assert "--b.arg_flag" in helptext


def test_subparsers_wrapping() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: int

    @dataclasses.dataclass
    class CheckoutCompletion:
        """Help message."""

        y: int

    help = get_helptext_with_checks(A | CheckoutCompletion)  # type: ignore
    # Both backends use full metavar when there's a single subparser group.
    # Appears 3 times: usage line + subcommand list.
    assert help.count("checkout-completion") == 2


def test_subparsers_wrapping1() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: int

    @dataclasses.dataclass
    class CheckoutCompletio:
        """Help message."""

        y: int

    help = get_helptext_with_checks(A | CheckoutCompletio)  # type: ignore
    # Both backends use full metavar when there's a single subparser group.
    assert help.count("checkout-completio") == 2


def test_subparsers_wrapping2() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: int

    @dataclasses.dataclass
    class CheckoutCompletionn:
        """Help message."""

        y: int

    help = get_helptext_with_checks(A | CheckoutCompletionn)  # type: ignore
    # Both backends use full metavar when there's a single subparser group.
    assert help.count("checkout-completionn") == 2


def test_subparsers_wrapping3() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: int

    @dataclasses.dataclass
    class CmdCheckout012:
        """Help message."""

        y: int

    help = get_helptext_with_checks(A | CmdCheckout012)  # type: ignore
    # Both backends use full metavar when there's a single subparser group.
    assert help.count("cmd-checkout012") == 2


def test_tuple_default() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: Tuple[str, str] = ("hello", "world")

    help = get_helptext_with_checks(A)
    assert "STR STR" in help
    assert "hello world" in help
    assert "('hello', 'world')" not in help


def test_argconf_constructor() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: Annotated[
            Tuple[str, str], tyro.conf.arg(constructor=lambda x: ("a", "b"))
        ] = ("hello", "world")

    help = get_helptext_with_checks(A)
    assert "STR STR" not in help
    # Unlike case above, should not be converted to 'hello world'.
    assert "('hello', 'world')" in help  # JSON special case.


def test_argconf_constructor_json_special_case() -> None:
    @dataclasses.dataclass
    class A:
        """Help message."""

        x: Annotated[
            # NOTE: this style of JSON construction is/will be deprecated.
            # Avoid.
            Tuple[str, str],
            tyro.conf.arg(constructor=json.loads, metavar="JSON"),
        ] = ("hello", "world")
        y: Annotated[
            # NOTE: this style of JSON construction is/will be deprecated.
            # Avoid.
            Tuple[int, int],
            tyro.conf.arg(constructor=json.loads, metavar="JSON"),
        ] = (3, 5)

    help = get_helptext_with_checks(A)
    assert "STR STR" not in help
    assert "JSON" in help
    assert '["hello", "world"]' in help  # JSON special case.
    assert "[3, 5]" in help, help  # JSON special case.


def test_optional_group() -> None:
    def f(
        x: Annotated[
            # NOTE: this style of JSON construction is/will be deprecated.
            # Avoid.
            Tuple[str, str],
            tyro.conf.arg(constructor=json.loads, metavar="JSON"),
        ] = ("hello", "world"),
        y: int = 3,
    ) -> int:
        del x, y
        return 5

    help = get_helptext_with_checks(f, default=3)
    assert 'default if used: ["hello", "world"]' in help
    assert "default if used: 3" in help


def test_append_fixed() -> None:
    def f(
        x: tyro.conf.UseAppendAction[Tuple[str, str]],
        y: int = 3,
    ) -> int:
        del x, y
        return 5

    help = get_helptext_with_checks(f)
    assert "repeatable" not in help, help


def test_append_good() -> None:
    def f(
        x: tyro.conf.UseAppendAction[Tuple[str, ...]],
        y: int = 3,
    ) -> int:
        del x, y
        return 5

    help = get_helptext_with_checks(f)
    assert "repeatable" in help, help


def test_append_with_default() -> None:
    def f(
        x: tyro.conf.UseAppendAction[Tuple[str, ...]] = ("hello world", "hello"),
        y: int = 3,
    ) -> int:
        del x, y
        return 5

    help = get_helptext_with_checks(f)
    assert "repeatable, appends to: 'hello world' hello" in help, help


def test_typeddict_exclude() -> None:
    class Special(TypedDict):
        x: NotRequired[int]

    help = get_helptext_with_checks(Special)
    assert "unset by default" in help, help


def test_conf_erased_argname() -> None:
    @dataclasses.dataclass(frozen=True)
    class Verbose:
        is_verbose: bool = True

    @dataclasses.dataclass(frozen=True)
    class Args:
        verbose: Annotated[Verbose, tyro.conf.arg(name="")]

    def main(args: Args) -> None:
        print(args)

    helptext = get_helptext_with_checks(main)
    assert "args options" in helptext
    assert "args.verbose options" not in helptext


def test_long_literal_options() -> None:
    """Test that long literal options don't get truncated in helptext."""

    def main(
        x: Literal[
            "one",
            "two",
            "three",
            "four",
            "five",
            "six",
            "seven",
            "eight",
            "nine",
            "ten",
            "eleven",
            "twelve",
            "thirteen",
            "fourteen",
            "fifteen",
            "sixteen",
            "seventeen",
            "eighteen",
            "nineteen",
            "twenty",
        ],
    ) -> None:
        print(x)

    helptext = get_helptext_with_checks(main)

    # Check that the helptext contains all options without truncation
    for option in [
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
        "nine",
        "ten",
        "eleven",
        "twelve",
        "thirteen",
        "fourteen",
        "fifteen",
        "sixteen",
        "seventeen",
        "eighteen",
        "nineteen",
        "twenty",
    ]:
        assert option in helptext

    # Check that there's no ellipsis character in the helptext
    assert "…" not in helptext


def test_bool_help_edge_cases() -> None:
    """Test edge cases in metavar generation for unions with tuples and literals."""

    # Use the exact same pattern as in bool_help_edge.py but as individual function definitions
    # Currently this passes, but represents edge cases that were problematic
    def main_int(x_int: Tuple[int] | Tuple[int, int] = (1,)) -> None:
        pass

    def main_literal(
        x_literal: Tuple[Literal[0, 1, 2]]
        | Tuple[Literal[0, 1, 2], Literal[0, 1, 2]] = (0,),
    ) -> None:
        pass

    def main_bool(x_bool: Tuple[bool] | Tuple[bool, bool] = (True,)) -> None:
        pass

    # Test int case - should show INT|{INT INT}
    helptext_int = get_helptext_with_checks(main_int)
    assert "--x-int INT|{INT INT}" in helptext_int

    # Test literal case - should show {0,1,2}|{{0,1,2} {0,1,2}}
    helptext_literal = get_helptext_with_checks(main_literal)
    assert "--x-literal {0,1,2}|{{0,1,2} {0,1,2}}" in helptext_literal

    # Test bool case - should show {True,False}|{{True,False} {True,False}}
    helptext_bool = get_helptext_with_checks(main_bool)
    assert "--x-bool {True,False}|{{True,False} {True,False}}" in helptext_bool


def test_set_accent_color() -> None:
    """Test that set_accent_color runs without error."""
    tyro.extras.set_accent_color("blue")


def test_help_with_required_subcommands_consolidated() -> None:
    """Test that --help shows help text for required subcommands in consolidated mode.

    This is a regression test for an issue where --help would raise a 'Missing subcommand'
    error instead of showing the help text when using CascadeSubcommandArgs with
    required subcommands (no default).
    """
    import pytest

    @dataclasses.dataclass
    class ConfigA:
        a: int = 1

    @dataclasses.dataclass
    class ConfigB:
        b: int = 2

    def main(config: ConfigA | ConfigB) -> None:
        """Test function with required union subcommand.

        Args:
            config: Configuration to use.
        """
        pass

    # This should show help text and exit, not raise a "Missing subcommand" error.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--help"], config=(tyro.conf.CascadeSubcommandArgs,))


def test_implicit_arguments_show_required() -> None:
    """Test that default subcommand options show '(required)' indicator for required args."""

    @dataclasses.dataclass
    class SubcommandWithRequiredArg:
        """Subcommand with a required argument."""

        required_arg: int  # Required: no default value.
        optional_arg: str = "default"  # Optional: has default value.

    @dataclasses.dataclass
    class SubcommandAllOptional:
        """Another subcommand with all optional args."""

        x: int = 0

    def main(
        subcommand: SubcommandWithRequiredArg | SubcommandAllOptional = (
            SubcommandWithRequiredArg(required_arg=tyro.MISSING)  # type: ignore
        ),
    ) -> None:
        """Test function with default subcommand containing required arg.

        Args:
            subcommand: Subcommand choice with default.
        """
        pass

    # Get help text with CascadeSubcommandArgs enabled.
    # The default subcommand's arguments should appear as "implicit arguments".
    helptext = get_helptext_with_checks(
        main,
        args=["--help"],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )

    # Implicit arguments only appear with tyro backend.
    if tyro._experimental_options["backend"] != "tyro":
        return

    # Check that default subcommand options section exists.
    assert "default subcommand options" in helptext

    # Check that required arg has "(required)" indicator.
    assert "--subcommand.required-arg" in helptext
    required_arg_pos = helptext.find("--subcommand.required-arg")
    assert required_arg_pos != -1
    # Look for "(required)" within a reasonable distance after the arg name.
    helptext_after_required = helptext[required_arg_pos : required_arg_pos + 150]
    assert "(required)" in helptext_after_required

    # Check that optional arg does NOT have "(required)" indicator.
    assert "--subcommand.optional-arg" in helptext
    # Find the line with the optional arg and verify it doesn't have "(required)".
    lines = helptext.split("\n")
    for line in lines:
        if "--subcommand.optional-arg" in line:
            assert "(required)" not in line
            break


def test_multiple_default_subcommands_with_required() -> None:
    """Test multiple default subcommands showing required indicators.

    This mirrors the pattern from examples/03_subcommands/03_multiple_subcommands.py
    where dataset and optimizer are both default subcommands with required fields.
    """

    @dataclasses.dataclass
    class Mnist:
        """MNIST dataset configuration."""

        binary: bool  # Required: no default value, like the example.

    @dataclasses.dataclass
    class ImageNet:
        """ImageNet dataset configuration."""

        subset: int = 1000

    @dataclasses.dataclass
    class Adam:
        """Adam optimizer configuration."""

        learning_rate: float = 1e-3

    @dataclasses.dataclass
    class Sgd:
        """SGD optimizer configuration."""

        learning_rate: float = 3e-4

    def train(
        dataset: Mnist | ImageNet = Mnist(tyro.MISSING),  # type: ignore
        optimizer: Adam | Sgd = Adam(),
    ) -> None:
        """Example training script.

        Args:
            dataset: Dataset to train on.
            optimizer: Optimizer to train with.
        """
        pass

    # Get help text with CascadeSubcommandArgs enabled.
    helptext = get_helptext_with_checks(
        train,
        args=["--help"],
        config=(tyro.conf.CascadeSubcommandArgs,),
    )

    # Implicit arguments only appear with tyro backend.
    if tyro._experimental_options["backend"] != "tyro":
        return

    # Check that default subcommand options section exists.
    assert "default subcommand options" in helptext

    # Check that the required boolean field from dataset (Mnist.binary) shows "(required)".
    assert "--dataset.binary" in helptext
    binary_pos = helptext.find("--dataset.binary")
    assert binary_pos != -1
    helptext_after_binary = helptext[binary_pos : binary_pos + 150]
    assert "(required)" in helptext_after_binary

    # Check that the optional optimizer field shows up but is NOT required.
    assert "--optimizer.learning-rate" in helptext
    lines = helptext.split("\n")
    for line in lines:
        if "--optimizer.learning-rate" in line:
            assert "(required)" not in line
            break


def test_compact_help_short_text() -> None:
    """Test compact mode with short argument text that fits on one line."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class ShortArgs:
        """Test program with short arguments."""

        x: int = 5
        """Short description."""

        y: str = "hello"
        """Another short one."""

    # Test compact mode.
    compact_helptext = get_helptext_with_checks(ShortArgs, compact_help=True)
    assert "--help-verbose" in compact_helptext
    assert "--x INT" in compact_helptext
    assert "--y STR" in compact_helptext
    # In compact mode, primary help text is omitted, only defaults shown.
    assert "Short description" not in compact_helptext
    assert "(default: 5)" in compact_helptext
    assert "(default: hello)" in compact_helptext


def test_compact_help_long_text() -> None:
    """Test compact mode with very long argument names and behavior hints."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class LongArgs:
        """Test program with very long argument names."""

        very_long_argument_name_that_takes_up_a_lot_of_terminal_space: int = 5
        """This is an extremely long description that will definitely exceed the terminal width when combined with the argument name. It contains many words and should trigger the line-splitting behavior in compact mode."""

    # Test compact mode.
    compact_helptext = get_helptext_with_checks(LongArgs, compact_help=True)
    assert (
        "--very-long-argument-name-that-takes-up-a-lot-of-terminal-space"
        in compact_helptext
    )
    # In compact mode, primary descriptions are omitted.
    assert "extremely long description" not in compact_helptext
    # But default values are still shown.
    assert "(default: 5)" in compact_helptext


def test_verbose_help_short_invocation() -> None:
    """Test verbose mode with short invocation names."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class VerboseArgs:
        """Test program for verbose mode."""

        x: int = 5
        """This is a longer description that will appear in verbose mode. It provides detailed information about what the x parameter does and how it should be used."""

        y: str = "test"
        """Another detailed description for the y parameter."""

    # Test verbose mode (default behavior).
    verbose_helptext = get_helptext_with_checks(VerboseArgs, compact_help=False)
    assert "--help-verbose" not in verbose_helptext
    assert "--x INT" in verbose_helptext
    assert "--y STR" in verbose_helptext
    assert "longer description" in verbose_helptext
    assert "Another detailed description" in verbose_helptext


def test_verbose_help_long_invocation() -> None:
    """Test verbose mode with very long invocation names that exceed max width."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class VeryLongInvocationArgs:
        """Test program with very long argument names."""

        extremely_long_argument_name_that_exceeds_maximum_width_threshold: int = 5
        """Description for the long argument."""

        another_incredibly_long_argument_name_for_testing_purposes: str = "test"
        """Another description."""

    # Test verbose mode with long invocations.
    verbose_helptext = get_helptext_with_checks(
        VeryLongInvocationArgs, compact_help=False
    )
    assert (
        "--extremely-long-argument-name-that-exceeds-maximum-width-threshold"
        in verbose_helptext
    )
    assert (
        "--another-incredibly-long-argument-name-for-testing-purposes"
        in verbose_helptext
    )
    assert "Description for the long argument" in verbose_helptext


def test_compact_vs_verbose_help() -> None:
    """Test that compact and verbose modes produce different output."""
    # Skip this test for argparse backend - compact_help only supported with tyro.
    if tyro._experimental_options["backend"] != "tyro":
        return

    @dataclasses.dataclass
    class CompareArgs:
        """Test comparing compact vs verbose."""

        x: int = 5
        """A description for parameter x."""

    compact = get_helptext_with_checks(CompareArgs, compact_help=True)
    verbose = get_helptext_with_checks(CompareArgs, compact_help=False)

    # Compact should have --help-verbose.
    assert "--help-verbose" in compact
    assert "--help-verbose" not in verbose

    # Both should have the basic argument.
    assert "--x INT" in compact
    assert "--x INT" in verbose


def test_no_duplicate_description() -> None:
    """Test that description passed via description= kwarg only appears once.

    Regression test for issue where description would appear both after usage line
    and inside the options box, causing duplication.
    """

    def main(x: int) -> None:
        pass

    # Test with description passed as keyword argument.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(
            main, args=["--help"], description="This helptext should only appear once."
        )

    helptext = tyro._strings.strip_ansi_sequences(target.getvalue())
    # The description should appear in the helptext.
    assert "This helptext should only appear once." in helptext
    # But it should only appear once, not twice.
    assert helptext.count("This helptext should only appear once.") == 1


def test_literal_invalid_choice_error_message() -> None:
    """Test that invalid choice error for Literal types shows clean argument name.

    Regression test for issue where internal implementation detail
    '__tyro-dummy-inner__' was exposed in error messages.
    """

    # When using a Literal type as the main CLI type, invalid choice errors
    # should show a clean argument name, not internal implementation details.
    # This will also be caught by the assertion in helptext_utils, but we test
    # it explicitly here for error messages.
    # Capture stderr to check the error message.
    stderr = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stderr(stderr):
            tyro.cli(Literal["a", "b"], args=["c"], return_unknown_args=True)  # type: ignore

    # Check the error message.
    error_message = stderr.getvalue()
    # The error message should mention the invalid choice 'c'.
    assert "invalid choice" in error_message.lower() or "c" in error_message
    # But it should NOT expose internal implementation details.
    assert "__tyro-dummy-inner__" not in error_message


def test_optional_union_subparser_helptext() -> None:
    """Test that optional union subparsers are marked as optional in help text."""

    @dataclasses.dataclass
    class StructA:
        """First option."""

        a: int = 1

    @dataclasses.dataclass
    class StructB:
        """Second option."""

        b: int = 2

    # Optional union (total=False).
    class OptionalConfig(TypedDict, total=False):
        """Config with optional union field."""

        choice: StructA | StructB
        other: int

    helptext_optional = get_helptext_with_checks(OptionalConfig)

    # Check that "(required)" is NOT shown for optional subparser.
    lines = helptext_optional.split("\n")
    in_subcommands_section = False
    for line in lines:
        if "subcommands" in line.lower():
            in_subcommands_section = True
        if in_subcommands_section and "choice:struct-a" in line:
            # The section header before the subcommand list should not say "(required)".
            assert (
                "(required)"
                not in helptext_optional.split("choice:struct-a")[0].split(
                    "subcommands"
                )[-1]
            )
            break

    # For tyro backend, check usage line shows optional subparser in brackets.
    # For argparse backend, the brackets may not show up in the usage line due to
    # argparse limitations, but "optional" should appear in the section title.
    if "[{choice:struct-a,choice:struct-b}]" not in helptext_optional:
        # Argparse backend - check for "optional" in section title.
        assert (
            "optional" in helptext_optional.lower()
            or "choice subcommands" in helptext_optional
        )

    # Required union (total=True by default).
    class RequiredConfig(TypedDict):
        """Config with required union field."""

        choice: StructA | StructB
        other: int

    helptext_required = get_helptext_with_checks(RequiredConfig)

    # Check usage line shows required subparser.
    assert "{choice:struct-a,choice:struct-b}" in helptext_required

    # Check that "(required)" IS shown for required subparser (tyro backend).
    # Argparse backend doesn't show "(required)" but doesn't mark subcommands as optional.
    if "(required)" not in helptext_required:
        # Argparse backend - ensure "optional subcommands" is NOT in the help text.
        # (The word "optional" may appear elsewhere, like in field descriptions)
        assert "optional subcommands" not in helptext_required.lower()


def test_subcommand_help_with_required_parent_args() -> None:
    """Test that subcommand --help works without satisfying parent required args.

    This is a regression test for https://github.com/brentyi/tyro/issues/403
    where using `python script.py sub:a --help` would fail with "Missing required
    argument: --x" instead of showing the subcommand help.
    """

    @dataclasses.dataclass
    class A:
        a: int

    @dataclasses.dataclass
    class B:
        b: int

    def main(x: int, sub: A | B) -> None:
        pass

    # This should show help text for the subcommand without requiring --x.
    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(main, args=["sub:a", "--help"])
    assert exc_info.value.code == 0

    # Verify the help text is actually shown (not an error).
    captured = io.StringIO()
    with contextlib.redirect_stdout(captured):
        with pytest.raises(SystemExit):
            tyro.cli(main, args=["sub:a", "--help"])
    helptext = captured.getvalue()
    assert "sub:a" in helptext or "--sub.a" in helptext

    # Also test with -h flag.
    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(main, args=["sub:a", "-h"])
    assert exc_info.value.code == 0

    # -H and --help-verbose are tyro-backend-only features, and only work
    # when compact_help=True.
    if tyro._experimental_options["backend"] == "tyro":
        # Test with compact_help=True: all help flags should work.
        with pytest.raises(SystemExit) as exc_info:
            tyro.cli(main, args=["sub:a", "--help"], compact_help=True)
        assert exc_info.value.code == 0

        with pytest.raises(SystemExit) as exc_info:
            tyro.cli(main, args=["sub:a", "-h"], compact_help=True)
        assert exc_info.value.code == 0

        with pytest.raises(SystemExit) as exc_info:
            tyro.cli(main, args=["sub:a", "-H"], compact_help=True)
        assert exc_info.value.code == 0

        with pytest.raises(SystemExit) as exc_info:
            tyro.cli(main, args=["sub:a", "--help-verbose"], compact_help=True)
        assert exc_info.value.code == 0



"""Tests for Unpack[TypedDict] support in **kwargs."""

from typing import NotRequired, Required, TypedDict, Unpack

import pytest

import tyro


class BasicOptions(TypedDict):
    """Basic TypedDict with required and optional fields."""

    option_a: str
    option_b: NotRequired[str]
    option_c: NotRequired[int]


class TotalFalseOptions(TypedDict, total=False):
    """TypedDict with total=False (all fields optional)."""

    option_a: str
    option_b: int


class MixedTotalFalseOptions(TypedDict, total=False):
    """TypedDict with total=False but some Required fields."""

    option_a: Required[str]
    option_b: int


def test_basic_unpack_kwargs() -> None:
    """Test basic Unpack[TypedDict] with required and optional fields."""

    def main(regular_arg: int, **kwargs: Unpack[BasicOptions]) -> dict:
        return {"regular": regular_arg, "kwargs": kwargs}

    # All fields provided.
    result = tyro.cli(
        main,
        args=[
            "--regular-arg",
            "42",
            "--kwargs.option-a",
            "hello",
            "--kwargs.option-b",
            "world",
            "--kwargs.option-c",
            "123",
        ],
    )
    assert result == {
        "regular": 42,
        "kwargs": {"option_a": "hello", "option_b": "world", "option_c": 123},
    }


def test_unpack_kwargs_required_only() -> None:
    """Test that only required field is mandatory."""

    def main(**kwargs: Unpack[BasicOptions]) -> dict:
        return dict(kwargs)

    # Only required field.
    result = tyro.cli(main, args=["--kwargs.option-a", "hello"])
    assert result == {"option_a": "hello"}


def test_unpack_kwargs_missing_required() -> None:
    """Test error when required field is missing."""

    def main(**kwargs: Unpack[BasicOptions]) -> dict:
        return dict(kwargs)

    with pytest.raises(SystemExit):
        tyro.cli(main, args=[])


def test_unpack_kwargs_total_false() -> None:
    """Test total=False TypedDict where all fields are optional."""

    def main(**kwargs: Unpack[TotalFalseOptions]) -> dict:
        return dict(kwargs)

    # No fields - should work.
    result = tyro.cli(main, args=[])
    assert result == {}

    # Some fields.
    result = tyro.cli(main, args=["--kwargs.option-a", "hello"])
    assert result == {"option_a": "hello"}


def test_unpack_kwargs_mixed_total_false() -> None:
    """Test total=False with Required[] fields."""

    def main(**kwargs: Unpack[MixedTotalFalseOptions]) -> dict:
        return dict(kwargs)

    # Required field only.
    result = tyro.cli(main, args=["--kwargs.option-a", "hello"])
    assert result == {"option_a": "hello"}

    # Missing required field should error.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=[])


def test_unpack_kwargs_with_class() -> None:
    """Test Unpack[TypedDict] in class __init__."""

    class MyClass:
        def __init__(self, name: str, **kwargs: Unpack[BasicOptions]) -> None:
            self.name = name
            self.kwargs = kwargs

    result = tyro.cli(MyClass, args=["--name", "test", "--kwargs.option-a", "value"])
    assert result.name == "test"
    assert result.kwargs == {"option_a": "value"}


def test_unpack_kwargs_with_args() -> None:
    """Test Unpack[TypedDict] combined with *args."""

    class MyClass:
        def __init__(self, *args: str, **kwargs: Unpack[BasicOptions]) -> None:
            self.args = args
            self.kwargs = kwargs

    result = tyro.cli(
        MyClass,
        args=["--args", "a", "b", "--kwargs.option-a", "hello"],
    )
    assert result.args == ("a", "b")
    assert result.kwargs == {"option_a": "hello"}


def test_plain_kwargs_still_works() -> None:
    """Test that plain **kwargs without Unpack still works."""

    def main(**kwargs: float) -> dict:
        return dict(kwargs)

    result = tyro.cli(main, args=["--kwargs", "learning_rate", "0.001", "beta", "0.9"])
    assert result == {"learning_rate": 0.001, "beta": 0.9}


def test_consistency_regular_vs_variadic_args() -> None:
    """Test consistent behavior between regular args and *args with Unpack[TypedDict].

    This test verifies that both FirstType (regular `args: str` parameter) and
    SecondType (`*args: str` variadic parameter) work correctly with
    `**kwargs: Unpack[TypedDict]`.
    """

    class Options(TypedDict):
        option_a: str
        option_b: NotRequired[str]

    class FirstType:
        """Class with regular args parameter and Unpack kwargs."""

        def __init__(self, args: str, **kwargs: Unpack[Options]) -> None:
            self.args = args
            self.kwargs = kwargs

    class SecondType:
        """Class with variadic *args and Unpack kwargs."""

        def __init__(self, *args: str, **kwargs: Unpack[Options]) -> None:
            self.args = args
            self.kwargs = kwargs

    # Test FirstType.
    first = tyro.cli(
        FirstType,
        args=["--args", "hello", "--kwargs.option-a", "value_a"],
    )
    assert first.args == "hello"
    assert first.kwargs == {"option_a": "value_a"}

    # Test SecondType.
    second = tyro.cli(
        SecondType,
        args=["--args", "hello", "--kwargs.option-a", "value_a"],
    )
    assert second.args == ("hello",)
    assert second.kwargs == {"option_a": "value_a"}

    # Test with optional kwargs field.
    first = tyro.cli(
        FirstType,
        args=[
            "--args",
            "hello",
            "--kwargs.option-a",
            "value_a",
            "--kwargs.option-b",
            "value_b",
        ],
    )
    assert first.args == "hello"
    assert first.kwargs == {"option_a": "value_a", "option_b": "value_b"}

    second = tyro.cli(
        SecondType,
        args=[
            "--args",
            "x",
            "y",
            "z",
            "--kwargs.option-a",
            "value_a",
            "--kwargs.option-b",
            "value_b",
        ],
    )
    assert second.args == ("x", "y", "z")
    assert second.kwargs == {"option_a": "value_a", "option_b": "value_b"}



"""Dictionaries and ``TypedDict``

Dictionary inputs can be specified using either a standard ``dict[K, V]``
annotation, or a :py:class:`TypedDict` subclass.

For configuring :py:class:`TypedDict`, we also support :code:`total={True/False}`,
:py:data:`typing.Required`, and :py:data:`typing.NotRequired`. See the `Python docs <https://docs.python.org/3/library/typing.html#typing.TypedDict>`_ for all :py:class:`TypedDict` features.

Usage:

    python ./04_dictionaries.py --help
    python ./04_dictionaries.py --typed-dict-a.learning-rate 3e-4 --typed-dict-b.betas 0.9 0.999
    python ./04_dictionaries.py --typed-dict-b.betas 0.9 0.999
"""

from typing import TypedDict

from typing_extensions import NotRequired

import tyro


class DictionarySchemaA(
    TypedDict,
    # Setting `total=False` specifies that not all keys need to exist.
    total=False,
):
    learning_rate: float
    betas: tuple[float, float]


class DictionarySchemaB(TypedDict):
    learning_rate: NotRequired[float]
    """NotRequired[] specifies that a particular key doesn't need to exist."""
    betas: tuple[float, float]


def main(
    typed_dict_a: DictionarySchemaA,
    typed_dict_b: DictionarySchemaB,
    standard_dict: dict[str, float] = {
        "learning_rate": 3e-4,
        "beta1": 0.9,
        "beta2": 0.999,
    },
) -> None:
    assert isinstance(typed_dict_a, dict)
    assert isinstance(typed_dict_b, dict)
    assert isinstance(standard_dict, dict)
    print("Typed dict A:", typed_dict_a)
    print("Typed dict B:", typed_dict_b)
    print("Standard dict:", standard_dict)


if __name__ == "__main__":
    tyro.cli(main)



"""Abstractions for pulling out 'field' definitions, which specify inputs, types, and # type: ignore
defaults, from general callables."""

from __future__ import annotations

import contextlib
import dataclasses
import functools
import inspect
import sys
from typing import Any, Callable, Dict, Literal, Tuple

import docstring_parser
from typing_extensions import (
    Annotated,
    Doc,
    get_args,
    get_origin,
    get_original_bases,
    is_typeddict,
)

from tyro.conf._mutex_group import _MutexGroupConfig
from tyro.constructors._primitive_spec import PrimitiveConstructorSpec

from . import _docstrings, _resolver, _strings, _unsafe_cache
from . import _fmtlib as fmt
from ._singleton import MISSING_NONPROP, is_missing
from ._typing import TypeForm
from ._typing_compat import is_typing_annotated, is_typing_unpack
from .conf import _confstruct, _markers
from .constructors._registry import ConstructorRegistry, check_default_instances
from .constructors._struct_spec import (
    InvalidDefaultInstanceError,
    StructFieldSpec,
    StructTypeInfo,
    UnsupportedStructTypeMessage,
)

global_context_markers: list[tuple[_markers.Marker, ...]] = []


@dataclasses.dataclass
class FieldDefinition:
    intern_name: str
    extern_name: str
    type: TypeForm[Any] | Callable
    """Full type, including runtime annotations."""
    type_stripped: TypeForm[Any] | Callable
    default: Any
    helptext: str | Callable[[], str | None] | None
    markers: set[Any]
    custom_constructor: bool

    argconf: _confstruct._ArgConfig
    mutex_group: _MutexGroupConfig | None

    # Override the name in our kwargs. Useful whenever the user-facing argument name
    # doesn't match the keyword expected by our callable.
    call_argname: Any

    # How this field should be passed to the callable.
    # - "kwarg": passed as keyword argument (default)
    # - "positional": passed as positional argument
    # - "unpack_args": unpacked as *args
    # - "unpack_kwargs": unpacked as **kwargs
    call_mode: Literal["kwarg", "positional", "unpack_args", "unpack_kwargs"] = "kwarg"

    @staticmethod
    @contextlib.contextmanager
    def marker_context(markers: tuple[_markers.Marker, ...]):
        """Context for setting markers on fields. All fields created within the
        context will have the specified markers."""
        global_context_markers.append(markers)
        try:
            yield
        finally:
            global_context_markers.pop()

    @staticmethod
    def from_field_spec(field_spec: StructFieldSpec) -> FieldDefinition:
        return FieldDefinition.make(
            name=field_spec.name,
            typ=field_spec.type,
            default=field_spec.default,
            helptext=field_spec.helptext,
            call_argname_override=field_spec._call_argname,
        )

    @staticmethod
    def make(
        name: str,
        typ: TypeForm[Any] | Callable,
        default: Any,
        helptext: str | Callable[[], str | None] | None,
        call_argname_override: Any | None = None,
        call_mode: Literal[
            "kwarg", "positional", "unpack_args", "unpack_kwargs"
        ] = "kwarg",
    ):
        # Narrow types.
        if typ is Any and not is_missing(default):
            typ = type(default)

        # Get all Annotated[] metadata.
        # This will unpack types in the form Annotated[type_stripped, *metadata].
        type_stripped, metadata = _resolver.unwrap_annotated(typ, search_type="all")

        # Support PEP 727 Doc objects.
        doc_objs = tuple(x for x in metadata if isinstance(x, Doc))
        if len(doc_objs) > 0:
            helptext = _strings.remove_single_line_breaks(
                _strings.dedent(doc_objs[0].documentation)
            ).strip()

        # Try to extract argconf overrides from type.
        argconfs = tuple(x for x in metadata if isinstance(x, _confstruct._ArgConfig))
        argconf = _confstruct._ArgConfig(
            None,
            None,
            help=None,
            help_behavior_hint=None,
            aliases=None,
            prefix_name=True,
            constructor_factory=None,
            default=MISSING_NONPROP,
        )
        for overwrite_argconf in argconfs:
            # Apply any annotated argument configuration values.
            update_values = {}
            for field in dataclasses.fields(overwrite_argconf):
                value = getattr(overwrite_argconf, field.name)
                # Handle default specially; we only want to apply it if it's
                # explicitly set (i.e., not MISSING_NONPROP).
                if field.name == "default":
                    if value is not MISSING_NONPROP:
                        update_values[field.name] = value
                elif value is not None:
                    update_values[field.name] = value

            argconf = dataclasses.replace(argconf, **update_values)
            if argconf.help is not None:
                helptext = argconf.help

        # Get markers.
        markers = tuple(x for x in metadata if isinstance(x, _markers._Marker))
        mutually_exclusive_groups = tuple(
            x for x in metadata if isinstance(x, _MutexGroupConfig)
        )

        # Include markers set via context manager.
        for context_markers in global_context_markers:
            markers += context_markers

        # Only use argconf default if field default is missing.
        if default is MISSING_NONPROP and len(argconfs) > 0:
            default = argconf.default

        # Construct field.
        out = FieldDefinition(
            intern_name=name,
            extern_name=name if argconf.name is None else argconf.name,
            type=typ,
            type_stripped=type_stripped,
            default=default,
            helptext=helptext,
            markers=set(markers),
            custom_constructor=argconf.constructor_factory is not None,
            argconf=argconf,
            mutex_group=mutually_exclusive_groups[0]
            if len(mutually_exclusive_groups) > 0
            else None,
            call_argname=(
                call_argname_override if call_argname_override is not None else name
            ),
            call_mode=call_mode,
        )

        # Be forgiving about default instances.
        type_stripped = _resolver.narrow_collection_types(type_stripped, default)
        if not check_default_instances():
            type_stripped = _resolver.expand_union_types(type_stripped, default)

        if type_stripped != out.type_stripped:
            return out.with_new_type_stripped(type_stripped)
        else:
            return out

    def with_new_type_stripped(
        self, new_type_stripped: TypeForm[Any] | Callable
    ) -> FieldDefinition:
        if is_typing_annotated(get_origin(self.type)):
            new_type = Annotated[(new_type_stripped, *get_args(self.type)[1:])]  # type: ignore
        else:
            new_type = new_type_stripped  # type: ignore
        return dataclasses.replace(
            self,
            type=new_type,  # type: ignore
            type_stripped=new_type_stripped,
        )


@_unsafe_cache.unsafe_cache(maxsize=1024)
def is_struct_type(
    typ: TypeForm[Any] | Callable, default_instance: Any, in_union_context: bool
) -> bool:
    """Determine whether a type should be treated as a 'struct type', where a single
    type can be broken down into multiple fields (eg for nested dataclasses or
    classes).

    The `in_union_context` flag indicates whether this type is being evaluated as part
    of a union. When True, allows collection types like List[Struct] or Dict[str, Struct]
    without defaults to be treated as struct types (for subcommand creation).
    """

    # Fast path: common primitive types are never struct types.
    if type(typ) is type and typ in (int, str, float, bool, bytes, type(None)):
        return False

    list_or_error = field_list_from_type_or_callable(
        typ,
        default_instance,
        support_single_arg_types=False,
        in_union_context=in_union_context,
    )
    return not isinstance(
        list_or_error,
        (UnsupportedStructTypeMessage, InvalidDefaultInstanceError),
    )


def field_list_from_type_or_callable(
    f: Callable | TypeForm[Any],
    default_instance: Any,
    support_single_arg_types: bool,
    in_union_context: bool,
) -> (
    UnsupportedStructTypeMessage
    | InvalidDefaultInstanceError
    | tuple[Callable | TypeForm[Any], list[FieldDefinition]]
):
    """Generate a list of generic 'field' objects corresponding to the inputs of some
    annotated callable.

    The `in_union_context` flag indicates whether this type is being evaluated as part
    of a union. When True, allows collection types like List[Struct] or Dict[str, Struct]
    without defaults to be treated as struct types (for subcommand creation).

    Returns:
        - tuple[type, list[FieldDefinition]] if successful: the resolved type and its field definitions.
        - UnsupportedStructTypeMessage if the type cannot be treated as a struct (e.g., not a dataclass, function, etc.).
        - InvalidDefaultInstanceError if the type can be treated as a struct, but the provided default instance is incompatible with the type.
    """
    if len(_resolver.unwrap_annotated(f, PrimitiveConstructorSpec)[1]) > 0:
        return UnsupportedStructTypeMessage(
            f"{f} should be parsed as a primitive type."
        )

    type_info = StructTypeInfo.make(f, default_instance, in_union_context)
    type_orig = f
    del f

    # Special case when treating `None` as a struct type.
    if support_single_arg_types and type_info.type is type(None):
        if not is_missing(default_instance) and default_instance is not None:
            return InvalidDefaultInstanceError(
                (
                    fmt.text(
                        "Default type ",
                        fmt.text["cyan"](str(type(default_instance))),
                        " is not ",
                        fmt.text["magenta"]("None"),
                    ),
                )
            )
        return (lambda: None, [])

    with type_info._typevar_context:
        spec = ConstructorRegistry.get_struct_spec(type_info)

        # Check if we got an error instead of a spec.
        if isinstance(spec, InvalidDefaultInstanceError):
            return spec

        with FieldDefinition.marker_context(type_info.markers):
            if spec is not None:
                return spec.instantiate, [
                    FieldDefinition.from_field_spec(f) for f in spec.fields
                ]

            is_primitive = ConstructorRegistry._is_primitive_type(type_orig, set())
            if is_primitive and support_single_arg_types:
                with FieldDefinition.marker_context((_markers.Positional,)):
                    return (
                        lambda x: x,
                        [
                            FieldDefinition.make(
                                name="value",
                                typ=type_orig,
                                default=default_instance,
                                helptext="",
                                call_mode="positional",
                            )
                        ],
                    )
            elif not is_primitive and callable(type_info.type):
                return _field_list_from_function(
                    type_info.type,  # This will have typing.Annotated metadata stripped.
                    default_instance,
                    markers=type_info.markers,
                )

    return UnsupportedStructTypeMessage(f"{type_orig} is not a valid struct type!")


def _field_list_from_function(
    f: Callable, default_instance: Any, markers: tuple[_markers.Marker, ...]
) -> UnsupportedStructTypeMessage | tuple[Callable, list[FieldDefinition]]:
    """Generate field lists from callables."""

    # Development note: separate conditions are helpful for test coverage reports.
    if f is Any:
        return UnsupportedStructTypeMessage("`Any` is not a valid struct type!")

    f_origin = get_origin(f)
    if getattr(f_origin, "__module__", None) in ("collections.abc", "builtins"):
        return UnsupportedStructTypeMessage(f"`{f}` is not a valid struct type!")

    try:
        params = list(inspect.signature(f).parameters.values())
    except ValueError:
        return UnsupportedStructTypeMessage(f"Could not get signature for {f}!")

    # `f` that is called (output) may be different from what we want to
    # inspect.
    f_out = f

    # Unwrap functools.wraps and functools.partial.
    done = False
    functools_marker = False
    while not done:
        done = True
        if hasattr(f, "__wrapped__"):
            f = f.__wrapped__  # type: ignore
            done = False
            functools_marker = True
        if isinstance(f, functools.partial):
            f = f.func
            done = False
            functools_marker = True

    # Check for abstract classes.
    if inspect.isabstract(f):
        return UnsupportedStructTypeMessage("Abstract classes cannot be instantiated!")

    # If `f` is class, we want to inspect its __init__ method for the
    # signature. But the docstrings may still be in the class signature itself.
    f_before_init_unwrap = f

    hints = None

    if inspect.isclass(f):
        signature_func = None
        if hasattr(f, "__init__") and f.__init__ is not object.__init__:
            signature_func = "__init__"
        elif hasattr(f, "__new__") and f.__new__ is not object.__new__:
            signature_func = "__new__"

        if signature_func is not None:
            # Get the __init__ / __new__ method from the class, as well as the
            # class that contains it.
            #
            # We call this the "signature function", because it's the function
            # that we use to instantiate the class.
            orig_cls = f
            base_cls_with_signature = None
            for base_cls_with_signature in inspect.getmro(f):
                if signature_func in base_cls_with_signature.__dict__:
                    f = getattr(base_cls_with_signature, signature_func)
                    break
            assert base_cls_with_signature is not None
            assert f is not orig_cls

            # For older versions of Python, the signature returned above (when
            # passed through generics base classes) will sometimes be (*args,
            # **kwargs).
            #
            # This is a hack. We can remove it if we deprecate Python 3.8 support.
            if sys.version_info < (3, 9) and not functools_marker:  # pragma: no cover
                params = list(inspect.signature(f).parameters.values())[1:]

            # Get hints for the signature function by recursing through the
            # inheritance tree. This is needed to correctly resolve type
            # parameters, which can be set anywhere between the input class and
            # the class where the __init__ or __new__ method is defined.
            def get_hints_for_signature_func(cls):
                typevar_context = _resolver.TypeParamResolver.get_assignment_context(
                    cls
                )
                cls = typevar_context.origin_type
                with typevar_context:
                    if cls is base_cls_with_signature:
                        return _resolver.get_type_hints_resolve_type_params(
                            f, include_extras=True
                        )
                    for base_cls in get_original_bases(cls):
                        if not issubclass(
                            _resolver.unwrap_origin_strip_extras(base_cls),
                            base_cls_with_signature,
                        ):
                            continue
                        return get_hints_for_signature_func(
                            _resolver.TypeParamResolver.resolve_params_and_aliases(
                                base_cls
                            )
                        )

                assert False, (
                    "We couldn't find the base class. This seems like a bug in tyro."
                )

            hints = get_hints_for_signature_func(orig_cls)

    # Early return for lambda functions.
    if getattr(f, "__name__", None) == "<lambda>" and len(params) > 0:
        return UnsupportedStructTypeMessage(
            "Lambda functions cannot be type-annotated!"
        )

    # Get type annotations, docstrings.
    docstring = inspect.getdoc(f)
    docstring_from_arg_name = {}
    if docstring is not None:
        for param_doc in docstring_parser.parse(docstring).params:
            docstring_from_arg_name[param_doc.arg_name] = param_doc.description
    del docstring

    # Get hints if we haven't done it already.
    # This will throw a type error for torch.device, typing.Dict, etc.
    if hints is None:
        try:
            hints = _resolver.get_type_hints_resolve_type_params(f, include_extras=True)
        except TypeError:
            return UnsupportedStructTypeMessage(f"Could not get hints for {f}!")

    # Expect non-empty type hints from classes.
    #
    # Generally we can be more forgiving with functions, for example
    #
    #     def main(x = 3) -> None: ...
    #
    # we can parse as if `x` was annotated with int. But if we have:
    #
    #     def main(x: SomeScaryType = SomeScaryDefault) -> None: ...
    #
    # we'll be more conservative in converting `--x` to a {fixed} argument.
    # The latter case requires returning an UnsupportedStructTypeMessage to avoid
    # unpacking the arguments of SomeScaryType.
    if (len(hints) == 0 or len(params) == 0) and not inspect.isfunction(
        f_before_init_unwrap
    ):
        return UnsupportedStructTypeMessage(f"Empty hints for {f}!")

    field_list = []
    for param in params:
        # Get default value.
        default = param.default

        # Get helptext from docstring.
        helptext = docstring_from_arg_name.get(param.name)
        if helptext is None and inspect.isclass(f_before_init_unwrap):
            # Lazy docstring parsing: use partial to defer expensive parsing.
            helptext = functools.partial(
                _docstrings.get_field_docstring,
                f_before_init_unwrap,
                param.name,
                markers,
            )

        # Set call_mode and markers for positional + variadic arguments.
        func_markers: Tuple[Any, ...] = ()
        call_mode: Literal["kwarg", "positional", "unpack_args", "unpack_kwargs"] = (
            "kwarg"
        )
        typ: Any = hints.get(param.name, Any)
        if param.kind is inspect.Parameter.POSITIONAL_ONLY:
            func_markers = (_markers.Positional,)
            call_mode = "positional"
        elif param.kind is inspect.Parameter.VAR_POSITIONAL:
            # Handle *args signatures.
            #
            # This will create a `--args T [T ...]` CLI argument.
            call_mode = "unpack_args"
            typ = Tuple[(typ, ...)]  # type: ignore
            # Only set empty default when there's no default_instance.
            # When default_instance is provided, we want MISSING_NONPROP so that
            # the _OPTIONAL_GROUP logic can return the default_instance directly.
            if is_missing(default_instance):
                default = ()
        elif param.kind is inspect.Parameter.VAR_KEYWORD:
            # Handle **kwargs signatures.
            call_mode = "unpack_kwargs"
            typ_origin = get_origin(typ)
            unpack_args = get_args(typ)

            # Check for Unpack[TypedDict] pattern.
            if (
                is_typing_unpack(typ_origin)
                and len(unpack_args) == 1
                and is_typeddict(unpack_args[0])
            ):
                # Treat as nested TypedDict struct.
                typ = unpack_args[0]
                default = MISSING_NONPROP  # Let TypedDict rule handle defaults.
            else:
                # Original behavior: creates `--kwargs STR T [STR T ...]` argument.
                typ = Dict[str, typ]  # type: ignore
                default = {}

        with FieldDefinition.marker_context(func_markers):
            field_list.append(
                FieldDefinition.make(
                    name=param.name,
                    # param.annotation doesn't resolve forward references.
                    typ=typ
                    if is_missing(default_instance)
                    else Annotated[(typ, _markers._OPTIONAL_GROUP)],  # type: ignore
                    default=default if default is not param.empty else MISSING_NONPROP,
                    helptext=helptext,
                    call_mode=call_mode,
                )
            )

    return f_out, field_list



"""Self-contained Python completion logic for embedding in shell scripts.

This module provides the completion code that will be embedded in bash/zsh scripts.
"""

from __future__ import annotations

import pathlib


def get_embedded_code() -> str:
    """Get the Python completion code for embedding in shell scripts.

    Reads the completion script from _completion_script.py and returns
    the source code ready to embed in a heredoc.

    Returns:
        Python code as a string, ready to embed in a heredoc.
    """
    script_path = pathlib.Path(__file__).parent / "_completion_script.py"
    source = script_path.read_text()

    # Extract only the necessary parts (skip module docstring at top).
    return source.partition('"""')[2].partition('"""')[2].strip()



from __future__ import annotations

import collections.abc
import dataclasses
import enum
import functools
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Sequence, Sized

from typing_extensions import cast, get_args, get_origin, is_typeddict

from tyro._typing_compat import is_typing_notrequired, is_typing_required
from tyro.constructors._primitive_spec import (
    PrimitiveTypeInfo,
    UnsupportedTypeAnnotationError,
)

from .. import _docstrings, _resolver
from .. import _fmtlib as fmt
from .._singleton import (
    EXCLUDE_FROM_CALL,
    MISSING,
    MISSING_NONPROP,
    is_missing,
)
from .._typing import TypeForm
from ..conf import _confstruct, _markers

if TYPE_CHECKING:
    from ._registry import ConstructorRegistry


@dataclasses.dataclass(frozen=True)
class UnsupportedStructTypeMessage:
    """Reason why a callable cannot be treated as a struct type."""

    message: str


@dataclasses.dataclass(frozen=True)
class InvalidDefaultInstanceError:
    """Return value when a default instance is not applicable to an annotated struct type."""

    message: tuple[fmt._Text, ...]


@dataclasses.dataclass(frozen=True)
class StructFieldSpec:
    """Behavior specification for a single field in our callable."""

    name: str
    """The name of the field. This will be used as a keyword argument for the
    struct's associated ``instantiate(**kwargs)`` function."""
    type: TypeForm
    """The type of the field. Can be either a primitive or a nested struct type."""
    default: Any
    """The default value of the field."""
    helptext: str | Callable[[], str | None] | None = None
    """Helpjext for the field."""
    # TODO: it's theoretically possible to override the argname with `None`.
    _call_argname: Any = None
    """Private: the name of the argument to pass to the callable. This is used
    for dictionary types."""
    is_default_overridden: None = None
    """Deprecated. No longer used."""


@dataclasses.dataclass(frozen=True)
class StructConstructorSpec:
    """Specification for a struct type, which is broken down into multiple
    fields.

    Each struct type is instantiated by calling an ``instantiate(**kwargs)``
    function with keyword a set of keyword arguments.

    Unlike :class:`PrimitiveConstructorSpec`, there is only one way to use this class.
    It must be returned by a rule in :class:`ConstructorRegistry`.
    """

    instantiate: Callable[..., Any]
    """Function to call to instantiate the struct."""
    fields: tuple[StructFieldSpec, ...]
    """Fields used to construct the callable. Each field is used as a keyword
    argument for the ``instantiate(**kwargs)`` function."""


@dataclasses.dataclass(frozen=True)
class StructTypeInfo:
    """Information used to generate constructors for struct types."""

    type: TypeForm
    """The type of the (potential) struct."""
    markers: tuple[Any, ...]
    """Markers from :mod:`tyro.conf` that are associated with this field."""
    default: Any
    """The default value of the struct, or a member of
    :data:`tyro.constructors.MISSING_AND_MISSING_NONPROP` if not present. In a
    function signature, this is ``X`` in ``def main(x=X): ...``. This can be
    useful for populating the default values of the struct."""
    _typevar_context: _resolver.TypeParamAssignmentContext
    in_union_context: bool
    """Flag indicating whether this type is being evaluated as part of a union.
    When True, allows collection types like List[Struct] or Dict[str, Struct]
    without defaults to be treated as struct types for subcommand creation."""

    @staticmethod
    def make(
        f: TypeForm | Callable, default: Any, in_union_context: bool
    ) -> StructTypeInfo:
        _, parent_markers = _resolver.unwrap_annotated(f, _markers._Marker)
        f, found_subcommand_configs = _resolver.unwrap_annotated(
            f, _confstruct._SubcommandConfig
        )

        # Apply default from subcommand config, but only if no default was passed in to `StructTypeInfo.make()`.
        #
        # If we have a subcommand that's annotated with:
        #
        #     x: (
        #       Annotated[SomeType1, subcommand(default=SomeType1("default1"))]
        #       | Annotated[SomeType2, subcommand(default=SomeType2("default2"))]
        #     ) = SomeType1("assignment1")
        #
        # The assigned default "assignment1" will be routed the `default`
        # argument of this function. The annotated defaults should be captured
        # in `found_subcommand_configs`.
        #
        # For the first subcommand, we should use the default from
        # "assignment1" and not "default1".
        #
        # We'll also use StructTypeInfo for default subcommand matching. This
        # won't work if we always overwrite the assigned default with the one
        # in the annotation.
        if is_missing(default) and len(found_subcommand_configs) > 0:
            default = found_subcommand_configs[0].default

        # Handle generics.
        typevar_context = _resolver.TypeParamResolver.get_assignment_context(f)
        f = typevar_context.origin_type
        f = _resolver.narrow_subtypes(f, default)
        f = _resolver.narrow_collection_types(f, default)

        return StructTypeInfo(
            cast(TypeForm, f),
            parent_markers,
            default,
            typevar_context,
            in_union_context,
        )


def apply_default_struct_rules(registry: ConstructorRegistry) -> None:
    """Apply default struct rules to the registry.

    This function registers all the struct rules for different types:
    - Dataclasses
    - TypedDict
    - Attrs classes
    - Dict
    - NamedTuple
    - Sequences
    - Tuples
    - Pydantic models
    """
    from .._fields import is_struct_type
    from ._struct_spec_attrs import attrs_rule
    from ._struct_spec_dataclass import dataclass_rule
    from ._struct_spec_ml_collections import ml_collections_rule
    from ._struct_spec_msgspec import msgspec_rule
    from ._struct_spec_pydantic import pydantic_rule

    # Register imported rules.
    registry.struct_rule(attrs_rule)
    registry.struct_rule(dataclass_rule)
    registry.struct_rule(ml_collections_rule)
    registry.struct_rule(msgspec_rule)
    registry.struct_rule(pydantic_rule)

    @registry.struct_rule
    def typeddict_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        # Is this a TypedDict?
        if not is_typeddict(info.type):
            return None

        cls = cast(type, info.type)

        # Handle TypedDicts.
        field_list = []
        valid_default_instance = (
            not is_missing(info.default) and info.default is not EXCLUDE_FROM_CALL
        )
        assert not valid_default_instance or isinstance(info.default, dict)
        total = getattr(cls, "__total__", True)
        assert isinstance(total, bool)
        assert not valid_default_instance or isinstance(info.default, dict)
        for name, typ in _resolver.get_type_hints_resolve_type_params(
            cls, include_extras=True
        ).items():
            typ_origin = get_origin(typ)

            # Unwrap Required[]/NotRequired[] early so we can check the inner type.
            if is_typing_required(typ_origin) or is_typing_notrequired(typ_origin):
                args = get_args(typ)
                assert len(args) == 1, (
                    "typing.Required[] and typing.NotRequired[T] require a concrete type T."
                )
                inner_typ = args[0]
                del args
            else:
                inner_typ = typ

            if valid_default_instance and name in cast(dict, info.default):
                default = cast(dict, info.default)[name]
            elif is_typing_required(typ_origin) and total is False:
                # Support total=False.
                default = MISSING
            elif total is False:
                # Support total=False.
                default = EXCLUDE_FROM_CALL
                if is_struct_type(inner_typ, MISSING_NONPROP, in_union_context=False):
                    # total=False behavior is unideal for nested structures.
                    pass
                    # raise _instantiators.UnsupportedTypeAnnotationError(
                    #     "`total=False` not supported for nested structures."
                    # )
            elif is_typing_notrequired(typ_origin):
                # Support typing.NotRequired[].
                default = EXCLUDE_FROM_CALL
            else:
                default = MISSING

            # Nested struct types need to be populated / can't be excluded from the call.
            # Note: Union types are NOT converted here - they create subparsers that
            # can be optional. When a union field has EXCLUDE_FROM_CALL as its default
            # (from TypedDict total=False or NotRequired[]), no subcommand needs to be
            # selected, and the field will be excluded from the result.
            if default is EXCLUDE_FROM_CALL and is_struct_type(
                inner_typ, MISSING_NONPROP, in_union_context=False
            ):
                default = MISSING_NONPROP

            field_list.append(
                StructFieldSpec(
                    name=name,
                    type=inner_typ,
                    default=default,
                    helptext=functools.partial(
                        _docstrings.get_field_docstring, cls, name, info.markers
                    ),
                )
            )
        return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))

    @registry.struct_rule
    def dict_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        origin = get_origin(info.type)
        args = get_args(info.type)
        if is_typeddict(info.type) or (
            info.type
            not in (
                Dict,
                dict,
                collections.abc.Mapping,
            )
            and origin
            not in (
                dict,
                collections.abc.Mapping,
            )
        ):
            return None

        # Check if we have a dict with struct values but no default.
        has_default = not is_missing(info.default)
        has_empty_default = has_default and len(info.default) == 0

        # No default provided or empty default.
        if not has_default or has_empty_default:
            # If the value type is not a primitive: we can try to treat as a struct.
            # This enables subcommands like `dict[str, SomeStruct] | SomeStruct2`.
            from ._registry import ConstructorRegistry

            if (
                origin in (dict, collections.abc.Mapping)
                and len(args) == 2
                and not ConstructorRegistry._is_primitive_type(
                    args[1], set(info.markers)
                )
            ):
                # Require a default (even an empty one) outside of union context.
                if not has_default and not info.in_union_context:
                    from .. import _fmtlib as fmt

                    raise UnsupportedTypeAnnotationError(
                        (
                            fmt.text(
                                "Type ",
                                fmt.text["cyan"](str(info.type)),
                                " with struct-type values requires a default value.",
                            ),
                        )
                    )
                # Allow empty defaults in union context, or when we have any default.
                return StructConstructorSpec(instantiate=dict, fields=())
            return None

        field_list = []
        for k, v in info.default.items():
            field_list.append(
                StructFieldSpec(
                    name=str(k) if not isinstance(k, enum.Enum) else k.name,
                    type=type(v),
                    default=v,
                    helptext=None,
                    _call_argname=k,
                )
            )
        return StructConstructorSpec(instantiate=dict, fields=tuple(field_list))

    @registry.struct_rule
    def namedtuple_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        if not _resolver.is_namedtuple(info.type):
            return None

        field_list = []
        field_defaults = getattr(info.type, "_field_defaults", {})
        field_names = getattr(info.type, "_fields", [])

        # Handle collections.namedtuple which doesn't have type annotations.
        type_hints = {field: Any for field in field_names}
        type_hints.update(
            _resolver.get_type_hints_resolve_type_params(info.type, include_extras=True)
        )
        for name, typ in type_hints.items():
            default = field_defaults.get(name, MISSING_NONPROP)

            if not is_missing(info.default) and hasattr(info.default, name):
                default = getattr(info.default, name)
            elif info.default is MISSING:
                default = MISSING

            field_list.append(
                StructFieldSpec(
                    name=name,
                    type=typ,  # type: ignore
                    default=default,
                    helptext=functools.partial(
                        _docstrings.get_field_docstring, info.type, name, info.markers
                    ),
                )
            )

        return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))

    @registry.struct_rule
    def variable_length_sequence_rule(
        info: StructTypeInfo,
    ) -> StructConstructorSpec | None:
        origin = get_origin(info.type)
        if origin not in (
            list,
            set,
            tuple,
            Sequence,
            collections.abc.Sequence,
        ):
            return None

        # Check if we have a collection with struct values but no default.
        has_default = not is_missing(info.default)
        has_empty_default = (
            has_default and isinstance(info.default, Sized) and len(info.default) == 0
        )

        # No default provided or empty default.
        if not has_default or has_empty_default:
            # If the contained type is not a primitive, we can try to treat as a struct.
            # This enables subcommands like `list[SomeStruct] | None`.
            args = get_args(info.type)
            if len(args) == 0:
                return None

            contained_type = cast(type, args[0])

            from ._registry import ConstructorRegistry

            if not ConstructorRegistry._is_primitive_type(
                contained_type, set(info.markers)
            ):
                # Contained type is not a primitive, so treat as struct.
                # Require a default (even an empty one) outside of union context.
                if not has_default and not info.in_union_context:
                    from .. import _fmtlib as fmt

                    raise UnsupportedTypeAnnotationError(
                        (
                            fmt.text(
                                "Type ",
                                fmt.text["cyan"](str(info.type)),
                                " with struct-type values requires a default value.",
                            ),
                        )
                    )

                # Allow empty defaults in union context, or when we have any default.
                if origin is tuple:
                    return StructConstructorSpec(instantiate=tuple, fields=())
                elif origin in (list, Sequence, collections.abc.Sequence):
                    return StructConstructorSpec(instantiate=list, fields=())
                elif origin is set:
                    return StructConstructorSpec(instantiate=set, fields=())
            return None

        # Default is not iterable or not empty - let the rest of the function handle it.
        if not isinstance(info.default, Iterable):
            return None

        # Cast is for mypy.
        contained_type = cast(
            type, get_args(info.type)[0] if get_args(info.type) else Any
        )

        # If the inner type is a primitive, we'll just treat the whole type as
        # a primitive.
        from ._registry import (
            ConstructorRegistry,
            PrimitiveConstructorSpec,
            PrimitiveTypeInfo,
        )

        contained_primitive_spec = ConstructorRegistry.get_primitive_spec(
            PrimitiveTypeInfo.make(contained_type, set(info.markers))
        )
        if (
            isinstance(contained_primitive_spec, PrimitiveConstructorSpec)
            # Why do we check nargs?
            # Because for primitives, we can't nest variable-length collections.
            #
            # For example, list[list[str]] can't be parsed as a single primitive.
            #
            # However, list[list[str]] can be parsed if the outer type is
            # handled as a struct (and a default value is provided, which we
            # check above).
            and contained_primitive_spec.nargs != "*"
        ):
            return None

        field_list = []
        for i, default_i in enumerate(info.default):
            field_list.append(
                StructFieldSpec(
                    name=str(i),
                    type=cast(type, contained_type),
                    default=default_i,
                    helptext="",
                )
            )

        return StructConstructorSpec(
            instantiate=type(info.default), fields=tuple(field_list)
        )

    @registry.struct_rule
    def tuple_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
        # It's important that this tuple rule is defined *after* the general sequence rule. It should take precedence.
        if info.type is not tuple and get_origin(info.type) is not tuple:
            return None

        # Fixed-length tuples.
        children = get_args(info.type)
        if Ellipsis in children:
            return None  # We don't handle variable-length tuples here

        # Infer more specific type when tuple annotation isn't subscripted.
        if len(children) == 0:
            if is_missing(info.default):
                return None
            else:
                assert isinstance(info.default, tuple)
                children = tuple(type(x) for x in info.default)

        if is_missing(info.default) or info.default is EXCLUDE_FROM_CALL:
            default_instance = (info.default,) * len(children)
        else:
            default_instance = info.default

        field_list: list[StructFieldSpec] = []
        for i, child in enumerate(children):
            default_i = default_instance[i]
            field_list.append(
                StructFieldSpec(
                    name=str(i),
                    type=child,
                    default=default_i,
                    helptext="",
                )
            )

        from ._registry import ConstructorRegistry

        # If the tuple only contains primitive types, we can just treat the
        # whole tuple as a primitive.
        #
        # We carve an exception when there are variable-length inner types, like
        # `tuple[list[int], list[str]]`.
        primitive_only = True
        for field in field_list:
            spec = ConstructorRegistry.get_primitive_spec(
                PrimitiveTypeInfo.make(field.type, set(info.markers))
            )
            if isinstance(spec, UnsupportedTypeAnnotationError) or spec.nargs == "*":
                primitive_only = False
                break

        if primitive_only:
            return None
        return StructConstructorSpec(instantiate=tuple, fields=tuple(field_list))



from __future__ import annotations

import argparse
import contextlib
import io
import os
from typing import Any, Callable

import pytest

import tyro
import tyro._fmtlib
import tyro._strings
from tyro._singleton import MISSING_NONPROP


def get_helptext_with_checks(
    f: Callable[..., Any],
    args: list[str] = ["--help"],
    use_underscores: bool = False,
    default: Any = MISSING_NONPROP,
    config: tuple[Any, ...] = (),
    compact_help: bool = False,
) -> str:
    """Get the helptext for a given tyro with input, while running various
    checks along the way."""
    # Should be empty.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(
            f,
            args=args,
            use_underscores=use_underscores,
            default=default,
            console_outputs=False,
            config=config,
            compact_help=compact_help,
        )
    assert target.getvalue() == "", target.getvalue()

    # Check tyro.extras.get_parser().
    parser = tyro.extras.get_parser(f, use_underscores=use_underscores)
    assert isinstance(parser, argparse.ArgumentParser)

    # Basic checks for completion scripts.
    with pytest.raises(SystemExit):
        tyro.cli(
            f,
            default=default,
            args=["--tyro-write-completion", "bash", os.devnull],
            config=config,
        )
    for shell in ["bash", "zsh"]:
        for command in ["--tyro-print-completion", "--tyro-write-completion"]:
            target = io.StringIO()
            with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
                if command == "--tyro-write-completion":
                    tyro.cli(
                        f, default=default, args=[command, shell, "-"], config=config
                    )
                else:
                    # `--tyro-print-completion` is deprecated! We should use `--tyro-write-completion` instead.
                    tyro.cli(f, default=default, args=[command, shell], config=config)
            output = target.getvalue()
            # Check that completion was generated (either by shtab or tyro).
            assert "shtab" in output or "tyro" in output

    # Test with underscores
    for shell in ["bash", "zsh"]:
        target = io.StringIO()
        with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
            tyro.cli(
                f,
                default=default,
                args=["--tyro_write_completion", shell, "-"],
                use_underscores=True,
                config=config,
            )
        output = target.getvalue()
        # Check that completion was generated (either by shtab or tyro).
        assert "shtab" in output or "tyro" in output

    # Get the actual helptext.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(
            f,
            args=args,
            use_underscores=use_underscores,
            default=default,
            config=config,
            compact_help=compact_help,
        )
    helptext = tyro._strings.strip_ansi_sequences(target.getvalue())

    # Assert that internal implementation details are not exposed in helptext.
    assert "__tyro-dummy-inner__" not in helptext, (
        "Internal implementation detail '__tyro-dummy-inner__' should not appear in helptext"
    )

    return helptext



"""Tests for completion spec generation and get_completions logic."""

import contextlib
import dataclasses
import io
import pathlib
from typing import Annotated, List, Literal

import pytest

import tyro


def _strip_comments_and_normalize(script: str) -> str:
    """Strip comments and normalize whitespace for comparison.

    This removes:
    - Lines starting with # (comments)
    - Empty lines
    - Leading/trailing whitespace

    Then normalizes remaining whitespace for comparison.
    """
    lines = []
    for line in script.split("\n"):
        stripped = line.strip()
        # Skip empty lines and comment lines.
        if not stripped or stripped.startswith("#"):
            continue
        lines.append(stripped)
    return "\n".join(lines)


def test_completion_output_parity() -> None:
    """Test that argparse and tyro backends produce equivalent completion scripts.

    The scripts should be functionally identical (ignoring comments and formatting).
    """

    def simple_function(
        x: int = 5,
        y: Annotated[str, tyro.conf.arg(aliases=["-y"])] = "hello",
        flag: bool = False,
    ) -> None:
        """Simple function for testing completion parity."""

    # Generate completion with argparse backend.
    original_backend = tyro._experimental_options["backend"]
    tyro._experimental_options["backend"] = "argparse"

    target_argparse = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target_argparse):
        tyro.cli(simple_function, args=["--tyro-print-completion", "bash"])

    # Generate completion with tyro backend.
    tyro._experimental_options["backend"] = "tyro"

    target_tyro = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target_tyro):
        tyro.cli(simple_function, args=["--tyro-print-completion", "bash"])

    # Restore original backend.
    tyro._experimental_options["backend"] = original_backend

    # Strip comments and compare.
    argparse_normalized = _strip_comments_and_normalize(target_argparse.getvalue())
    tyro_normalized = _strip_comments_and_normalize(target_tyro.getvalue())

    # Different architectures but same functionality.
    # Argparse backend uses bash variables, tyro backend uses embedded Python.
    assert "_option_strings=" in argparse_normalized
    assert "COMPLETION_SPEC" in tyro_normalized
    assert "PYTHON_EOF" in tyro_normalized

    # Check that both have the same options (though in possibly different order/format).
    assert "--x" in argparse_normalized and "--x" in tyro_normalized
    assert "-y" in argparse_normalized and "-y" in tyro_normalized
    assert "--flag" in argparse_normalized and "--flag" in tyro_normalized
    assert "--no-flag" in argparse_normalized and "--no-flag" in tyro_normalized


def test_completion_parity_with_subcommands() -> None:
    """Test completion parity for simple subcommands (no CascadeSubcommandArgs)."""

    @dataclasses.dataclass
    class SubA:
        value_a: int = 1

    @dataclasses.dataclass
    class SubB:
        value_b: str = "hello"

    @dataclasses.dataclass
    class Config:
        sub: SubA | SubB

    # Generate completion with argparse backend.
    original_backend = tyro._experimental_options["backend"]
    tyro._experimental_options["backend"] = "argparse"

    target_argparse = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target_argparse):
        tyro.cli(Config, args=["--tyro-print-completion", "bash"])

    # Generate completion with tyro backend.
    tyro._experimental_options["backend"] = "tyro"

    target_tyro = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target_tyro):
        tyro.cli(Config, args=["--tyro-print-completion", "bash"])

    # Restore original backend.
    tyro._experimental_options["backend"] = original_backend

    argparse_output = target_argparse.getvalue()
    tyro_output = target_tyro.getvalue()

    # Both should have subcommands.
    assert "sub:sub-a" in argparse_output or "sub_a" in argparse_output
    assert "sub:sub-b" in argparse_output or "sub_b" in argparse_output
    assert "sub:sub-a" in tyro_output or "sub_a" in tyro_output
    assert "sub:sub-b" in tyro_output or "sub_b" in tyro_output

    # Both should have the subcommand arguments (possibly with prefix).
    assert (
        "--value-a" in argparse_output
        or "--value_a" in argparse_output
        or "--sub.value-a" in argparse_output
        or "--sub.value_a" in argparse_output
    )
    assert (
        "--value-b" in argparse_output
        or "--value_b" in argparse_output
        or "--sub.value-b" in argparse_output
        or "--sub.value_b" in argparse_output
    )
    assert (
        "--value-a" in tyro_output
        or "--value_a" in tyro_output
        or "--sub.value-a" in tyro_output
        or "--sub.value_a" in tyro_output
    )
    assert (
        "--value-b" in tyro_output
        or "--value_b" in tyro_output
        or "--sub.value-b" in tyro_output
        or "--sub.value_b" in tyro_output
    )


def test_path_completion(backend: str) -> None:
    """Test that path arguments get proper file/directory completion."""
    from pathlib import Path

    def path_function(
        input_file: Path,
        output_dir: Path,
        config_path: str,
    ) -> None:
        """Function with path arguments."""

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(path_function, args=["--tyro-print-completion", "bash"])

    output = target.getvalue()

    if backend == "tyro":
        # New Python-based backend: check for path types in the spec.
        assert "'type': 'path'" in output
    else:
        # Argparse backend uses shtab helper functions.
        assert "_shtab_compgen_files" in output or "_shtab_compgen_dirs" in output


def test_cascaded_subcommand_completion(backend: str) -> None:
    """Test completion for cascaded subcommands.

    This tests the tyro-specific CascadeSubcommandArgs feature, which allows
    more flexible argument ordering. The completion should include all available
    options at each level.
    """

    @dataclasses.dataclass
    class SubA:
        sub_value: int = 1

    @dataclasses.dataclass
    class SubB:
        sub_value: int = 2

    @dataclasses.dataclass
    class Config:
        parent_arg: str = "default"
        sub: SubA | SubB = dataclasses.field(default_factory=SubA)

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(
            tyro.conf.CascadeSubcommandArgs[Config],
            args=["--tyro-print-completion", "bash"],
        )

    output = target.getvalue()

    # Both backends should generate completions with subcommands.
    assert "sub:sub-a" in output or "sub_a" in output
    assert "sub:sub-b" in output or "sub_b" in output
    # Parent argument should be available.
    assert "--parent-arg" in output or "--parent_arg" in output


def test_cascade_marker_detection(backend: str) -> None:
    """Test that CascadeSubcommandArgs marker is properly detected in completion spec."""
    if backend != "tyro":
        pytest.skip("Cascade marker detection is tyro-specific")

    @dataclasses.dataclass
    class Config:
        regular_field: int = 5
        cascade_field: tyro.conf.CascadeSubcommandArgs[str] = "default"

    # Generate completion script.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(Config, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()

    # The completion spec is embedded in the script.
    # Verify it contains the fields and cascade info.
    assert "regular-field" in completion_script or "regular_field" in completion_script
    assert "cascade-field" in completion_script or "cascade_field" in completion_script

    # Verify the spec has cascade markers.
    # The cascade field should be tracked in the spec.
    assert "'cascade'" in completion_script


def test_nargs_with_choices_completion(backend: str) -> None:
    """Test that nargs is properly tracked for choice options in completion spec."""
    if backend != "tyro":
        pytest.skip("Choice nargs tracking is tyro-specific")

    @dataclasses.dataclass
    class Config:
        # Single choice.
        mode: Literal["train", "eval"] = "train"
        # Multiple choices.
        modes: List[Literal["train", "eval", "test"]] = dataclasses.field(
            default_factory=lambda: ["train"]
        )

    # Generate completion script.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(Config, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()

    # Verify completion spec has choices and nargs.
    assert "'choices'" in completion_script
    assert "train" in completion_script
    assert "eval" in completion_script
    assert "test" in completion_script

    # Verify nargs is tracked for the list field.
    assert "'nargs'" in completion_script


def test_metavar_in_description(backend: str) -> None:
    """Test that metavar is included in option descriptions."""
    if backend != "tyro":
        pytest.skip("Metavar formatting is tyro-specific")

    @dataclasses.dataclass
    class Config:
        count: int = 5
        name: str = "default"

    # Generate completion script.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(Config, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()

    # Verify metavar (type hints) appear in descriptions.
    # The descriptions should include INT and STR metavars.
    assert "INT" in completion_script or "int" in completion_script
    assert "STR" in completion_script or "str" in completion_script


@dataclasses.dataclass
class _ConfigForBulletTest:
    # No custom helptext, just default.
    simple: int = 5
    # Custom helptext.
    documented: int = 10
    """This is a custom help message."""


def test_smart_bullet_separator(backend: str) -> None:
    """Test that bullet separator is only used when there's custom helptext."""
    if backend != "tyro":
        pytest.skip("Bullet separator logic is tyro-specific")

    # Generate completion script.
    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(_ConfigForBulletTest, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()

    # Verify bullet separator is used for fields with custom helptext.
    # The custom help message should have the bullet.
    assert "custom help" in completion_script.lower()
    assert "\u2022" in completion_script


@dataclasses.dataclass
class _PreprocessingA:
    scale: float = 1.0


@dataclasses.dataclass
class _PreprocessingB:
    normalize: bool = True


@dataclasses.dataclass
class _AugmentationA:
    rotate: bool = True


@dataclasses.dataclass
class _AugmentationB:
    flip: bool = False


@dataclasses.dataclass
class _DatasetMNIST:
    """Dataset with its own frontier groups."""

    preprocessing: _PreprocessingA | _PreprocessingB
    augmentation: _AugmentationA | _AugmentationB


@dataclasses.dataclass
class _DatasetImageNet:
    resolution: int = 224


@dataclasses.dataclass
class _SubcommandWithFrontier:
    """A subcommand that has subcommands, one of which has frontier groups."""

    dataset: _DatasetMNIST | _DatasetImageNet


@dataclasses.dataclass
class _NestedMainCommand:
    # This creates a subcommand that has frontier groups.
    config: _SubcommandWithFrontier


def test_nested_subcommand_coverage(backend: str) -> None:
    """Test nested subcommands for coverage of _build_subcommand_spec recursion."""
    if backend != "tyro":
        pytest.skip("Testing tyro-specific completion generation")

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(_NestedMainCommand, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()
    # Verify nested subcommands are present.
    assert "config" in completion_script
    # Check for deeply nested frontier groups within subcommands.
    assert "dataset" in completion_script or "preprocessing" in completion_script


def test_positional_argument_coverage(backend: str) -> None:
    """Test that positional arguments are skipped in completion spec."""
    if backend != "tyro":
        pytest.skip("Testing tyro-specific completion generation")

    def main(input_file: tyro.conf.Positional[str], output: str = "out.txt") -> str:
        """Test with positional argument.

        Args:
            input_file: Input file path (positional).
            output: Output file path.
        """
        return f"{input_file} -> {output}"

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(main, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()
    # Positional args shouldn't appear in options list, only regular flags.
    assert "--output" in completion_script


def test_count_action_coverage(backend: str) -> None:
    """Test count action type for coverage."""
    if backend != "tyro":
        pytest.skip("Testing tyro-specific completion generation")

    def main(verbose: tyro.conf.UseCounterAction[int]) -> int:
        """Test function with counter action.

        Args:
            verbose: Verbosity level.
        """
        return verbose

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(main, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()
    # Count action creates a flag-type option.
    assert "flag" in completion_script.lower() or "verbose" in completion_script


def test_metavar_and_helptext_edge_cases(backend: str) -> None:
    """Test that completion works even with minimal help text."""
    if backend != "tyro":
        pytest.skip("Testing tyro-specific completion generation")

    def main(value: int = 5, flag: bool = False) -> int:
        """Test completion with basic arguments."""
        return value

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(main, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()
    # Verify basic completion generation works.
    assert "value" in completion_script or "flag" in completion_script


def test_unsupported_shell_error(backend: str) -> None:
    """Test that unsupported shell types raise an error."""
    if backend != "tyro":
        pytest.skip("Testing tyro-specific error handling")

    def main(value: int = 5) -> int:
        """Test function."""
        return value

    # Test that tcsh (supported by argparse but not tyro) raises ValueError.
    with pytest.raises(ValueError, match="Unsupported shell.*tcsh"):
        tyro.cli(main, args=["--tyro-print-completion", "tcsh"])


def test_nested_dataclass_completion(backend: str) -> None:
    """Test that nested dataclass fields are included in completion spec.

    This tests the case where a dataclass has a nested dataclass field (not a Union),
    which creates child parsers via child_from_prefix rather than subcommands.
    """

    @dataclasses.dataclass
    class OptimizerConfig:
        learning_rate: float = 3e-4
        weight_decay: float = 1e-2

    @dataclasses.dataclass
    class Config:
        opt: OptimizerConfig
        seed: int = 0

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stdout(target):
        tyro.cli(Config, args=["--tyro-print-completion", "bash"])

    completion_script = target.getvalue()

    # Verify that the top-level argument is present.
    assert "--seed" in completion_script

    # Verify that nested arguments are present with their prefix.
    assert "--opt.learning-rate" in completion_script
    assert "--opt.weight-decay" in completion_script


# Unit tests for reconstruct_colon_words.


def test_reconstruct_colon_words_basic() -> None:
    """Test basic word reconstruction for colon-separated subcommands."""
    from typing import Any, Dict

    from tyro._backends._completion._completion_script import reconstruct_colon_words

    # Spec with colon-separated subcommands.
    spec: Dict[str, Any] = {
        "subcommands": {
            "dataset:mnist": {},
            "dataset:image-net": {},
            "optimizer:adam": {},
        }
    }

    # Test 1: Basic reconstruction of "dataset:mnist".
    # Bash splits this as ["dataset", ":", "mnist"].
    words = ["dataset", ":", "mnist"]
    reconstructed, new_cword = reconstruct_colon_words(words, 2, spec)
    assert reconstructed == ["dataset:mnist"]
    assert new_cword == 0

    # Test 2: Partial completion "dataset:".
    # Bash splits this as ["dataset", ":"].
    words = ["dataset", ":"]
    reconstructed, new_cword = reconstruct_colon_words(words, 1, spec)
    assert reconstructed == ["dataset:"]
    assert new_cword == 0

    # Test 3: Partial prefix "dataset:m".
    # Bash splits this as ["dataset", ":", "m"].
    words = ["dataset", ":", "m"]
    reconstructed, new_cword = reconstruct_colon_words(words, 2, spec)
    assert reconstructed == ["dataset:m"]
    assert new_cword == 0


def test_reconstruct_colon_words_no_match() -> None:
    """Test that words are not reconstructed when they don't match subcommands."""
    from typing import Any, Dict

    from tyro._backends._completion._completion_script import reconstruct_colon_words

    spec: Dict[str, Any] = {"subcommands": {"dataset:mnist": {}}}

    # Test 1: Non-matching colon pattern (e.g., option value "key:value").
    # Should not be merged since "key:value" is not a known subcommand.
    words = ["--config", "key", ":", "value"]
    reconstructed, new_cword = reconstruct_colon_words(words, 3, spec)
    # Since "key:value" doesn't match any subcommand, keep them separate.
    # Standalone colons are skipped.
    assert reconstructed == ["--config", "key", "value"]
    assert new_cword == 2


def test_reconstruct_colon_words_multiple() -> None:
    """Test reconstruction with multiple colon-separated subcommands."""
    from typing import Any, Dict

    from tyro._backends._completion._completion_script import reconstruct_colon_words

    spec: Dict[str, Any] = {
        "subcommands": {
            "dataset:mnist": {},
            "optimizer:adam": {},
        }
    }

    # Test: Multiple subcommands "dataset:mnist optimizer:adam".
    # Bash splits as ["dataset", ":", "mnist", "optimizer", ":", "adam"].
    words = ["dataset", ":", "mnist", "optimizer", ":", "adam"]
    reconstructed, new_cword = reconstruct_colon_words(words, 5, spec)
    assert reconstructed == ["dataset:mnist", "optimizer:adam"]
    assert (
        new_cword == 1
    )  # Cursor on "adam" -> after reconstruction it's on second word.


def test_reconstruct_colon_words_with_options() -> None:
    """Test reconstruction with flags mixed in."""
    from typing import Any, Dict

    from tyro._backends._completion._completion_script import reconstruct_colon_words

    spec: Dict[str, Any] = {
        "subcommands": {
            "dataset:mnist": {},
        }
    }

    # Test: Subcommand with options "dataset:mnist --lr 0.001".
    # Bash splits as ["dataset", ":", "mnist", "--lr", "0.001"].
    words = ["dataset", ":", "mnist", "--lr", "0.001"]
    reconstructed, new_cword = reconstruct_colon_words(words, 4, spec)
    assert reconstructed == ["dataset:mnist", "--lr", "0.001"]
    assert new_cword == 2  # Cursor on "0.001" -> after merging, it's at index 2.


def test_reconstruct_colon_words_cursor_on_colon() -> None:
    """Test cursor position when it's on the colon itself."""
    from typing import Any, Dict

    from tyro._backends._completion._completion_script import reconstruct_colon_words

    spec: Dict[str, Any] = {"subcommands": {"dataset:mnist": {}}}

    # Test: Cursor is on the colon character.
    # Words: ["dataset", ":"], cursor at index 1 (the colon).
    words = ["dataset", ":"]
    reconstructed, new_cword = reconstruct_colon_words(words, 1, spec)
    # The colon should be merged with "dataset" to form "dataset:".
    assert reconstructed == ["dataset:"]
    # Cursor should be on the merged word.
    assert new_cword == 0


def test_reconstruct_colon_words_empty() -> None:
    """Test reconstruction with empty word list."""
    from typing import Any, Dict

    from tyro._backends._completion._completion_script import reconstruct_colon_words

    spec: Dict[str, Any] = {"subcommands": {}}

    # Empty word list.
    words: list[str] = []
    reconstructed, new_cword = reconstruct_colon_words(words, 0, spec)
    assert reconstructed == []
    assert new_cword == 0


# Unit tests for get_completions.


def test_get_completions_path_marker() -> None:
    """Test that get_completions returns path marker for string/Path arguments."""
    from typing import Any, Dict, cast

    from tyro._backends._completion._completion_script import get_completions
    from tyro._backends._completion._spec import build_completion_spec
    from tyro._parsers import ParserSpecification
    from tyro._singleton import MISSING_NONPROP

    @dataclasses.dataclass
    class Config:
        input_file: pathlib.Path = pathlib.Path("input.txt")
        name: str = "default"
        count: int = 0
        verbose: bool = False

    # Build spec from dataclass.
    parser_spec = ParserSpecification.from_callable_or_type(
        Config,
        markers=set(),
        description=None,
        parent_classes=set(),
        default_instance=MISSING_NONPROP,
        intern_prefix="",
        extern_prefix="",
        subcommand_prefix="",
        support_single_arg_types=False,
        prog_suffix="",
    )
    # Cast needed because get_completions uses Dict[str, Any] (it's standalone).
    spec = cast(Dict[str, Any], build_completion_spec(parser_spec, "prog"))

    # Test 1: Completing after --input-file should return path marker.
    completions = get_completions(["prog", "--input-file", ""], 2, spec)
    assert completions == ["__TYRO_COMPLETE_FILES__"]

    # Test 2: Completing after --name (str) should also return path marker.
    completions = get_completions(["prog", "--name", ""], 2, spec)
    assert completions == ["__TYRO_COMPLETE_FILES__"]

    # Test 3: Completing after --count (int) should return empty.
    completions = get_completions(["prog", "--count", ""], 2, spec)
    assert completions == []

    # Test 4: Completing after --verbose (flag) should show all options.
    completions = get_completions(["prog", "--verbose", ""], 2, spec)
    assert len(completions) > 0
    completion_flags = [c.split("\t")[0] for c in completions]
    assert "--input-file" in completion_flags
    assert "--name" in completion_flags
    assert "--count" in completion_flags

    # Test 5: Completing at root level should show all options.
    completions = get_completions(["prog", ""], 1, spec)
    completion_flags = [c.split("\t")[0] for c in completions]
    assert "--input-file" in completion_flags
    assert "--name" in completion_flags
    assert "--count" in completion_flags
    assert "--verbose" in completion_flags



"""ML collections + field references

``ml_collections`` supports references for sharing values across multiple fields.

Usage:
    python ./09_ml_collections_refs.py --help
    python ./09_ml_collections_refs.py --config.hidden-dim 32
    python ./09_ml_collections_refs.py --config.network.policy-hidden-dim 64
"""

from pprint import pprint

from ml_collections import ConfigDict, FieldReference  # type: ignore

import tyro


def get_config() -> ConfigDict:
    config = ConfigDict()

    # Placeholder.
    hidden_dim_ref = FieldReference(128)
    config.hidden_dim = hidden_dim_ref

    # Wandb config.
    config.wandb = ConfigDict()
    config.wandb.mode = "online"  # online, offline, disabled.
    config.wandb.project = "robot-sandbox"

    # Network config.
    # Updating `policy_hidden_dim` will update `value_hidden_dim`, but
    # updating `value_hidden_dim` will not update `policy_hidden_dim`.
    config.network = ConfigDict()
    config.network.policy_hidden_dim = hidden_dim_ref
    config.network.value_hidden_dim = hidden_dim_ref * 2
    config.network.policy_obs_key = "state"
    config.network.value_obs_key = "state"

    return config


def train(config: ConfigDict = get_config()) -> None:
    """Train a model."""
    pprint(config.to_dict())  # type: ignore


if __name__ == "__main__":
    tyro.cli(train)
